import os
from math import radians

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches
import matplotlib.animation

from gonioanalysis.directories import ANALYSES_SAVEDIR, CODE_ROOTDIR

from .common import vector_plot, surface_plot, make_animation_timestep

CURRENT_ARROW_LENGTH = 1


def error_at_flight(manalyser):
    '''
    Plot 2D heatmap with vertical limits (upper_lower) as x,y axes and
    pixel color values as the on flight orientation.

    To see what (lower,upper) vertical excluding would give smallest error.
    '''

    #optimal_rot = 50
    optimal_rot = 50

    rotations = np.linspace(-180, 180, 180)

    widths = np.linspace(10, 200, 20)
    centers = np.linspace(-50, 50, 10)

    # Absolute minimum errors, no matter where happens
    min_errors = [] 
    # Distance of the minimum error from the flight position
    distances = []

    for i, width in enumerate(widths):
        print('upper {}/{}'.format(i+1, len(widths)))
        min_errors.append([])
        distances.append([])
        for center in centers:
            manalyser.set_angle_limits(va_limits=(-width/2+center, width/2+center))
            p, e = optic_flow_error(manalyser, rotations)
            e = np.mean(e, axis=1)
            
            
            error_argmin = np.argmin(e)
            distance = rotations[error_argmin] - optimal_rot

            distances[-1].append(distance)
            min_errors[-1].append(np.min(e))

    manalyser.set_angle_limits()

    for im, figname, text in zip([min_errors, distances], ['min_errors.jpg', 'distances.jpg'], ['Minimum error (between 0 and 1)','The distance of flight rotation from minimum error (degrees)']):
        
        fig = plt.figure()
        zlimit = max(abs(np.min(im)), abs(np.max(im)))
        plt.imshow(im, extent=[np.min(centers), np.max(centers), np.min(widths), np.max(widths)], cmap='seismic',vmin=-zlimit, vmax=zlimit, origin='lower')
        plt.xlabel('Center point (degrees)')
        plt.ylabel('Width (degrees)')
        cbar = plt.colorbar()
        cbar.set_label(text)

        savedir = os.path.join(ANALYSES_SAVEDIR, 'error_at_flight', manalyser.get_specimen_name())
        os.makedirs(savedir, exist_ok=True)
        fig.savefig(os.path.join(savedir,figname))
        plt.close()



def complete_flow_analysis(manalyser, rotations, rotation_axis,
        text=True, animate=True, subdivide_regions=False, mutation_scale=3,
        dual_difference=True, elevations=[50, 0, -50]):
    '''
    Creates the combined plot of the rhabdomere movement vectormap, simulated
    optic flow and the difference between these two, rendered from 3 views
    
    PLOT
    ROW
    1       measured optic flow vectors
    2       simulated optic flow
    3       error heatplot

    INPUT ARGUMENTS
    manalyser
    rotations
    rotation_axis       Either 'yaw', 'roll', or 'pitch'
    text                wheter to render text
    dual_difference     Create a a difference plot for the opposite direction
                        as well, adding a column in the plot
    subdivide_regions   True/False, will divide the 
    

    '''
    from scipy.ndimage import rotate
    import matplotlib.image as mpli 
    from gonioanalysis.coordinates import rotate_vectors, optimal_sampling
    from gonioanalysis.drosom.optic_flow import flow_vectors, field_error

    # Parse keywork aguments
    # -----------------------
    N_fig_columns = 4
    if dual_difference:
        N_fig_columns += 1
    N_fig_rows = 3
    
    if not text:
        global CURRENT_ARROW_LENGTH
        CURRENT_ARROW_LENGTH = 1.5

    if text:
        dpi = 150
    else:
        dpi = 600

    elevation_texts = ['Dorsal\nview', 'Anterior\nview', 'Ventral\nview']

    column_texts = ['Biphasic receptive field\nmovement directions', #'Pseudopupil movement\ndirections',
            'Experienced optic flow',
            'Difference with slower phase',
            'Head orientation']
    
    if dual_difference:
        column_texts.insert(3, 'Difference with fast phase')

    if rotation_axis == 'pitch':
        zero_rot = 10
        optimal_rot = 10
        fly_image = os.path.join(CODE_ROOTDIR, 'droso6_rotated.png')
        sideflow=True
    elif rotation_axis == 'yaw':
        zero_rot = 0
        optimal_rot = 0
        fly_image = os.path.join(CODE_ROOTDIR, 'rotation_yaw.png')
        sideflow=True
    elif rotation_axis == 'roll':
        zero_rot = 0
        optimal_rot = 0
        fly_image = os.path.join(CODE_ROOTDIR, 'rotation_roll.png')
        sideflow=False


    # End parsing keyword arguments
    # ------------------------------

    lp, lvecs = manalyser.get_3d_vectors('left')
    rp, rvecs = manalyser.get_3d_vectors('right')
    
    points = np.concatenate((lp, rp))
    lrvecs = np.concatenate((lvecs, rvecs))

    # Flow field errors for each rotation
    #points, all_errors = optic_flow_error(manalyser, rotations)
    
    # Flow field vectors for each rotation
    vector_points = optimal_sampling(np.arange(-90, 90, 5), np.arange(-180, 180, 5))

    if rotation_axis == 'yaw':
        all_flow_vectors = [rotate_vectors(vector_points, flow_vectors(vector_points), -radians(rot), 0, 0) for rot in rotations]
    elif rotation_axis == 'pitch':
        all_flow_vectors = [rotate_vectors(vector_points, flow_vectors(vector_points), 0, -radians(rot), 0) for rot in rotations]
    elif rotation_axis == 'roll':
        all_flow_vectors = [rotate_vectors(vector_points, flow_vectors(vector_points), 0, 0, -radians(rot)) for rot in rotations] 

    all_errors = [field_error(points, lrvecs, *flow_vectors) for flow_vectors in all_flow_vectors]
    
    # 1D errorplot for the mean error over rotations
    #average_errors_1D = np.mean(all_errors, axis=1)
    #average_errors_1D_stds = np.std(all_errors, axis=1)
    
    
    savedir = os.path.join(ANALYSES_SAVEDIR, 'comparision_to_optic_flow', manalyser.get_specimen_name()+'_'+rotation_axis)
    os.makedirs(savedir, exist_ok=True)
    
    # SUBDIVIDE REGIONS
    # ------------------
    
    subdivide_flow_points = optimal_sampling(np.arange(-90, 90, 3), np.arange(-180, 180, 3))
    
    if subdivide_regions:
        subdivide_regions = [(-70, 70), (70, None), (None, -70)]
    else:
        subdivide_regions = [(None, None)]

    subdivide_regions_colors = ['black', 'gray', 'gray']#['pink', 'lime', 'turquoise']
    subdivide_styles = ['-', '-', '--']
    subdivide_lws = [3, 1, 1]
    subdivide_errors = []
    
    subdivide_points = []
    subdivide_vectors = []
    subdivide_flow_vectors = []

    for reg in subdivide_regions:
        manalyser.set_angle_limits(va_limits=reg)
        
        reglp, reglvecs = manalyser.get_3d_vectors('left')
        regrp, regrvecs = manalyser.get_3d_vectors('right')
        
        regpoints = np.concatenate((reglp, regrp))
        reglrvecs = np.concatenate((reglvecs, regrvecs))

        if rotation_axis == 'yaw':
            reg_all_flow_vectors = [rotate_vectors(subdivide_flow_points, flow_vectors(subdivide_flow_points), -radians(rot), 0, 0) for rot in rotations]
        elif rotation_axis == 'pitch':
            reg_all_flow_vectors = [rotate_vectors(subdivide_flow_points, flow_vectors(subdivide_flow_points), 0, -radians(rot), 0) for rot in rotations]
        elif rotation_axis == 'roll':
            reg_all_flow_vectors = [rotate_vectors(subdivide_flow_points, flow_vectors(subdivide_flow_points), 0, 0, -radians(rot)) for rot in rotations] 


        reg_all_errors = [field_error(regpoints, reglrvecs, *flow_vectors) for flow_vectors in reg_all_flow_vectors]
       
        #p, e = optic_flow_error(manalyser, all_flow_errors)
        e = np.mean(reg_all_errors, axis=1)
        subdivide_errors.append(e)

        #subdivide_flow_vectors.append( [flow_vectors(p, xrot=rot) for rot in rotations] )
        subdivide_flow_vectors.append( reg_all_flow_vectors )
        
        subdivide_points.append([reglp, regrp])
        subdivide_vectors.append([reglvecs, regrvecs])

    manalyser.set_angle_limits(va_limits=(None, None))
   
    # END OF SUBDIVIDE REGIONS
    # -------------------------

   
    
    fim = mpli.imread(fly_image)
    im_scaler = len(fim)
    

    N_steady_frames = 20*3
    steadied = False

    i_image = -1  # Saved image number
    i_steady = 0   # Steady looping
    i_rot = -1   # Determines current rotation
    while True:
        
        i_rot += 1
        rot = rotations[i_rot]

        print('Rotation {} degrees'.format(rot))

        if i_steady > N_steady_frames:
            steadied = True

        if rot > optimal_rot and not steadied:
            i_rot -= 1
            i_steady += 1
        
        if not text:
            if rot < optimal_rot:
                continue


        # Data collection part
        flow_vectors = all_flow_vectors[i_rot]
        flow_errors = all_errors[i_rot]

        i_image += 1
        fn = 'image_{:0>8}.png'.format(i_image)
        savefn = os.path.join(savedir, fn)
        
  
        # Plotting part        
        fig = plt.figure(figsize=(11.69,8.27), dpi=dpi)   
        
       

        for i, elev in enumerate(elevations):
            ax = fig.add_subplot(N_fig_rows, N_fig_columns, N_fig_columns*i+1, projection='3d')

            for ppp, vecs, color in zip(subdivide_points, subdivide_vectors, subdivide_regions_colors): 

                lcolor = 'red'
                rcolor = 'blue'

                if color != 'black':
                    lcolor = color
                    rcolor = color

                vector_plot(ax, ppp[0], vecs[0], color=lcolor, mutation_scale=mutation_scale, animate=animate, camerapos=(elev,90))
                vector_plot(ax, ppp[1], vecs[1], color=rcolor, mutation_scale=mutation_scale, animate=animate, camerapos=(elev,90))
            
            ax.view_init(elev=elev, azim=90)
            
            ax.dist = 6
            
            if text:
                if dual_difference:
                    ax.text2D(0, 0.5, elevation_texts[i].replace('\n', ' '), transform=ax.transAxes, va='center', ha='center', rotation=90)
                else:
                    ax.text2D(-0.15, 0.5, elevation_texts[i], transform=ax.transAxes, va='center')

        


    


        for i, elev in enumerate(elevations):
            ax = fig.add_subplot(N_fig_rows, N_fig_columns, N_fig_columns*i+2, projection='3d')
            vector_plot(ax, *flow_vectors, color='darkviolet', mutation_scale=mutation_scale, camerapos=(elev,90))
            ax.view_init(elev=elev, azim=90)

            ax.dist = 6

        for i, elev in enumerate(elevations):
            ax = fig.add_subplot(N_fig_rows, N_fig_columns, N_fig_columns*i+3, projection='3d')
            m = surface_plot(ax, points, 1-np.array(flow_errors), cb=False)
            ax.view_init(elev=elev, azim=90)

            ax.dist = 6
        
        if dual_difference:
            for i, elev in enumerate(elevations):
                ax = fig.add_subplot(N_fig_rows, N_fig_columns, N_fig_columns*i+4, projection='3d')
                m = surface_plot(ax, points, np.array(flow_errors), cb=False)
                ax.view_init(elev=elev, azim=90)

                ax.dist = 6
            
        
        axes = fig.get_axes()
        j=0
        for ikk, ax in enumerate(axes):
            if ikk in [0,3,6,9]:
                if text:
                    ax.text2D(0.5, 1.05, column_texts[j], transform=ax.transAxes, ha='center', va='top')
                j +=1

        # Plot image of the fly
        ax = fig.add_subplot(N_fig_rows, N_fig_columns, N_fig_columns)
        ax.imshow(rotate(fim, rot, mode='nearest', reshape=False), cmap='gray')
        ax.set_frame_on(False)


        # Stop video at the optimal
        if 0<i_steady<N_steady_frames:
            rect = matplotlib.patches.Rectangle((0,0), 1, 1, transform=ax.transAxes, fill=False,
                    color='yellow', linewidth=8)
            if text:
                ax.add_patch(rect)
                ax.text(0.5, 1.1, 'Typical head tilt', ha='center', va='top', transform=ax.transAxes)

        # Add arrows
        if text:
            if sideflow == True:
                arrows = [np.array((1.05,ik))*im_scaler for ik in np.arange(0.1,0.91,0.1)]
                for x, y in arrows:
                    plt.arrow(x, y, -0.1*im_scaler, 0, width=0.01*im_scaler, color='darkviolet')
            else:
                x = []
                y = []
                for i in range(10):
                    for j in range(10):
                        x.append((0.5+i)*fim.shape[0]/10)
                        y.append((0.5+j)*fim.shape[1]/10)
                ax.scatter(x,y, marker='x', color='darkviolet')

        # Text
        ax = fig.add_subplot(N_fig_rows, N_fig_columns, 2*N_fig_columns)
        cbox = ax.get_position()
        print(cbox)
        cbox.x1 -= abs((cbox.x1 - cbox.x0))/1.1
        cbox.y0 -= 0.18*abs(cbox.y1-cbox.y0)
        cax = fig.add_axes(cbox)
        plt.colorbar(m, cax)
        if text:
            if dual_difference:
                text_x = 0.1
            else:
                text_x = 0.025
            cax.text(text_x, 0.95, 'Matching', va='top', transform=ax.transAxes)
            cax.text(text_x, 0.5, 'Perpendicular', va='center', transform=ax.transAxes)
            cax.text(text_x, 0.05, 'Opposing', va='bottom', transform=ax.transAxes)


        # Axis for the error
        ax = fig.add_subplot(N_fig_rows, N_fig_columns, 3*N_fig_columns)        
        ax_pos = ax.get_position()
        ax_pos = [ax_pos.x0+0.035, ax_pos.y0-0.08, ax_pos.width+0.022, ax_pos.height+0.05]
        ax.remove()
        ax = fig.add_axes(ax_pos)


        if text:
            ax.text(0.6, 1, 'Head tilt\n{} degrees'.format(int(rot)-zero_rot), transform=ax.transAxes, va='bottom', ha='left')
        
        
        
        # Plot the last figure, showing the mean error etc.
        if subdivide_errors:


            for subdivide_errors_1D, color, style, lw in zip(subdivide_errors, subdivide_regions_colors, subdivide_styles, subdivide_lws):
                if not text and color != 'black':
                    continue
                ax.plot(rotations-zero_rot, subdivide_errors_1D, style, lw=lw, color=color, label='Slower phase')
                # "Cursor"
                ax.scatter(rotations[i_rot]-zero_rot, subdivide_errors[0][i_rot], color='black')
                
                if dual_difference:
                    ax.plot(rotations-zero_rot, 1-subdivide_errors_1D, style, lw=lw, color='gray', label='Fast phase')
                    ax.scatter(rotations[i_rot]-zero_rot, 1-subdivide_errors[0][i_rot], color='gray')

            # Make optimal band
            ax.legend(loc=(0.39,1.2))
            left_side = -50
            right_side = 50
            p = matplotlib.patches.Rectangle((left_side, 0), right_side-left_side, 1, alpha=0.5, color='yellow')
            ax.add_patch(p)
            
            if text:
                ax.set_ylabel('Mean error')
            
            ax.set_ylim(np.min(subdivide_errors)-0.05, np.max(subdivide_errors)+0.05)
            l = matplotlib.lines.Line2D([rot-zero_rot, rot-zero_rot], [np.min(subdivide_errors)-0.05, np.max(subdivide_errors)+0.05], color='black')
            
            if text:
                ax.add_line(l)
                  
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        
        
        ax.set_xlim(-180-zero_rot, 180-zero_rot)
        
        
        if text:
            ax.set_yticks([0.2, 0.4, 0.6, 0.8])
            ax.set_xticks([-90, 0, 90])
            ax.set_xticklabels(['-90$^\circ$', '0$^\circ$','90$^\circ$'])
        else:
            ax.set_xticks([-90, 0, 90])
            ax.set_xticklabels([])
            ax.set_yticks([0.2, 0.4, 0.6, 0.8])
            ax.set_yticklabels([])


        # Hide axis from all the figures that are 3D plots or plots of images
        axes = fig.get_axes()
        for ax in axes[0:-1]:
            ax.set_axis_off()
        
        if len(subdivide_regions) == 3:
            # Plot subdivide_regions 
            for i_ax, ax in enumerate(axes[0:8]):
                reg = subdivide_regions[0]
                color = subdivide_regions_colors[0]
                for theta in reg:
                    

                    style = '-'
                    if theta < 0 and i_ax in [0,3]:
                        style = '--'
                    elif theta > 0 and i_ax in [2, 5]:
                        style = '--'
                     
                    
                    phi = np.linspace(0,180)
                    phi = np.radians(phi)
                    theta = np.radians(90-theta)
                    x = np.cos(phi)
                    y = np.sin(phi) * np.sin(theta)
                    
                    z = np.sin(phi) * np.cos(theta)

                   
                    ax.plot(x,y,z, style, color=color, lw=1)
        

        if dual_difference:
            plt.subplots_adjust(left=0.02, bottom=0.05, right=0.95, top=0.90, wspace=0.0, hspace=0.1)
        else:
            plt.subplots_adjust(left=0.08, bottom=0, right=0.95, top=0.95, wspace=0.1, hspace=0.1)
        
        make_animation_timestep(step_size=0.025, low_val=0.7)

        if savefn:
            fig.savefig(savefn)
        plt.close()




'''Alternative 3D plotting using hosguibase's render3d
'''

import os
import math
import tkinter as tk

import numpy as np
from panda3d.core import (
        NodePath,
        DirectionalLight,
        AmbientLight,
        Filename,
        )
import devjoni.guibase as gb
from devjoni.hosguibase.render3d import SceneWidget, SceneObject

from gonioanalysis.directories import CODE_ROOTDIR
from gonioanalysis.coordinates import rotate_along_arbitrary


named_colors = {
        'white': (1,1,1,1),
        'black': (0,0,0,1),
        'red': (1,0,0,1),
        'blue': (0,0,1,1),
        'green': (0,1,0,1),
        'orange': (1,0.5,0,1),
        'pink': (1,0.75,0.75,1),
        'gray': (0.5,0.5,0.5,1),
        'magenta': (1,0,1,1),
        }


class WrapTk():
    def __init__(self, tkwidget):
        self.tk = tkwidget


class RotateButtons(gb.FrameWidget):
    '''Rotate buttons
    '''
    def __init__(self, parent, obj, use_buttons=False):
        super().__init__(parent)
        self.parent = parent
        self.obj = obj

        if use_buttons:
            self.horp = gb.ButtonWidget(self, 'h+',
                                        command=self.rhorp)
            self.horp.grid(row=0,column=0)
            self.horm = gb.ButtonWidget(self, 'h-',
                                        command=self.rhorm)
            self.horm.grid(row=0,column=1)

            self.verp = gb.ButtonWidget(self, 'v+',
                                        command=self.rverp)
            self.verp.grid(row=0,column=2)
            self.verm = gb.ButtonWidget(self, 'v-',
                                        command=self.rverm)
            self.verm.grid(row=0,column=3)
        else:
            self.hor_slider = gb.SliderWidget(
                    self, from_=-90, to=90)
            self.hor_slider.grid(row=0, column=0, sticky='WE',
                                 row_weight=0)
            self.hor_slider.set_command(self.set_hor)

            self.ver_slider = gb.SliderWidget(
                    self, from_=-90, to=90)
            self.ver_slider.grid(row=1, column=0, sticky='WE',
                                 row_weight=0)
            self.ver_slider.set_command(self.set_ver)


    def set_ver(self, value):
        rot = self.obj.get_hpr()
        self.obj.set_hpr(rot[0], float(value), rot[2])
        self.parent.update()

    def set_hor(self, value):
        rot = self.obj.get_hpr()
        self.obj.set_hpr(float(value)+180, rot[1], rot[2])
        self.parent.update()

    def rhorp(self):
        rot = self.obj.get_hpr()
        self.obj.set_hpr(rot[0]+10, rot[1], rot[2])
        self.parent.update()
    
    def rhorm(self):
        rot = self.obj.get_hpr()
        self.obj.set_hpr(rot[0]-10, rot[1], rot[2]) 
        self.parent.update()

    def rverp(self):
        rot = self.obj.get_hpr()
        self.obj.set_hpr(rot[0], rot[1]+10, rot[2])
        self.parent.update()

    def rverm(self):
        rot = self.obj.get_hpr()
        self.obj.set_hpr(rot[0], rot[1]-10, rot[2])
        self.parent.update()


class Cp3d(gb.FrameWidget):
    '''Use in place of CanvasPlotter
    '''

    def __init__(self, parent):
        if isinstance(parent, tk.Frame):
            parent = WrapTk(parent)
        super().__init__(parent)

        self.ax3d = Ax3d(self)
        self.ax3d.grid(row=0, column=0)
        
        self.obj = SceneObject('central')
        self.obj.scene = self.ax3d.scene

        self.ax3d.nodeparent = self.obj
        self.ax3d.nodeparent.set_hpr(180,0,0)

        self.buttons = RotateButtons(self, self.obj)
        self.buttons.grid(row=1, column=0)

    def get_figax(self):
        return self.ax3d, self.ax3d

    def update(self):
        self.ax3d.scene.render()

class Ax3d(gb.FrameWidget):
    '''Use in place of matplotlib ax.
    '''

    def __init__(self, parent, nodeparent=None):
        super().__init__(parent)

        self.nodeparent = nodeparent

        self.scene = SceneWidget(self)
        self.scene.grid()

        self.scene.camera.set_pos(0,-9,0)
        self.scene.camera.set_hpr(0,0,0)
        

        self._arrow = SceneObject('arrow')
        self._arrow.scene = self.scene
        self._arrow.load_model(
                Filename.fromOsSpecific(
                    os.path.join(CODE_ROOTDIR, 'progdata', 'arrow.egg')
                    )
                )
        self._arrow.np.set_scale(0.05)
        self._arrow.np.setTwoSided(True)
        self._arrow.np.set_hpr(180,-90,0)
        self._arrow.np.flatten_strong()
        self._arrow.np.detachNode()

        self.arrows = []
       
    

        # Head
        self.head = SceneObject('head')
        self.head.scene = self.scene
        self.head.load_model(
                Filename.fromOsSpecific(
                    os.path.join(CODE_ROOTDIR, 'progdata', 'head.egg')
                    )
                )
        self.head.np.detachNode()
        self.head.np.setScale(3.3)
        self.head.np.setColorScale((0.8,0.8,0.8,1))
        #self.head.np.setShaderAuto()
        self._enable_lights()

        self.scene.showbase.setBackgroundColor((1,1,1,1))

    def show_head(self):
        self.head.np.reparentTo(self.nodeparent.np)

    def hide_head(self):
        self.head.np.detachNode()

    def _enable_lights(self):
        render = self.scene._rendernp
        #render.setShaderAuto()

        ambient = AmbientLight('ambient')
        ambient.setColor((0.3, 0.3, 0.3, 1))
        ambient_np = render.attachNewNode(ambient)
        render.setLight(ambient_np)
        
        for p in [0,1]:
            pitch = -30+p*60
            for heading in [0+p*60,120+p*60,240+p*60]:
                lamp = DirectionalLight(f'lamp-h{heading}-p{p}')
                lamp.setColor((0.4+p*0.2, 0.4+p*0.2, 0.4+p*0.2, 1))
                lamp_np = render.attachNewNode(lamp)
                lamp_np.setHpr(heading, pitch, 0)
                render.setLight(lamp_np)

    def set_axis_off(self):
        pass

    def set_xlabel(self, label):
        pass

    def set_ylabel(self, label):
        pass

    def set_zlabel(self, label):
        pass

    def add_arrow(self, x0,y0,z0,x1,y1,z1,
                  color=None, mutation_scale=6):

        arrow = NodePath('arrow')
        self._arrow.np.instance_to(arrow)

        arrow.reparent_to(self.nodeparent.np)

        h = math.degrees(math.atan2(y0,x0))+90

        dxy0 = math.sqrt(x0**2+y0**2)
        p = math.degrees(math.atan2(z0,dxy0))

        # Calculate roll by rotating the arrow to (0,1,0)
        p0 = np.array([(x0,y0,z0)])
        p1 = np.array([(x1,y1,z1)])
        
        p0 = rotate_along_arbitrary((0,0,1), p0, math.radians(h))[0]
        p1 = rotate_along_arbitrary((1,0,0), p1, math.radians(p))[0]
        
        pdx = p1[0]-p0[0]
        pdz = p1[2]-p0[2]
        r = math.degrees(math.atan2(pdx,pdz))

        arrow.set_hpr(h,-(p),-r+180) 

        if x0 < 0:
            shift = -0.6
        else:
            shift = 0.6 

        nx0 = x0*0.8+shift
        ny0 = y0*0.7
        nz0 = z0*0.8
     
        arrow.set_pos(nx0, ny0, nz0)
        colscale = named_colors.get(color, (0.9,0.9,0.9))
        arrow.setColor(colscale)  

        self.arrows.append(arrow)

    def plot_surface(self, X, Y, Z, color=None):
        pass

    def plot(self, X, Y, Z):
        pass

    def clear(self):
        
        for arrow in self.arrows:
            arrow.removeNode()

        self.arrows = []
        
        self.hide_head()

    def view_init(self, elev, azim):
        '''
        elev, azim : float
            Rotations in degrees
        '''
        self.nodeparent.set_hpr(-azim-90,-elev,0)
        self.show_head()

    @property
    def azim(self):
        return -self.nodeparent.get_hpr()[0]-90

    @azim.setter
    def azim(self, value):
        h,p,r = self.nodeparent.get_hpr()
        self.nodeparent.set_hpr(-value-90, p, r)
    
    @property
    def elev(self):
        return -self.nodeparent.get_hpr()[1]

    @elev.setter
    def elev(self, value):
        h,p,r = self.nodeparent.get_hpr()
        return self.nodeparent.set_hpr(h, value, r)





import os
import subprocess
import sys
import platform

from gonioanalysis.droso import SpecimenGroups
from gonioanalysis.directories import CODE_ROOTDIR
from gonioanalysis.drosom.analysers import analyser_classes
from gonioanalysis.directories import ANALYSES_SAVEDIR
from gonioanalysis.drosom.virtual_analyser import VirtualAnalyser

class Core:
    '''
    Tkinter independent functions, reusable for other GUI implementations.
    
    Attributes
    ----------
    data_directory : list of strings
        Current data directories
    current_specimen : string
        Name of the current specimen
    analyser : object
        MAnalyser (or OAnalsyer) object of the current specimen
    selected_recording : string
        Selected recording name (image_folder)
    analyser_class : class
        Class of the new analysers to create (MAnalyser or OAnalyser)
    analyser_classes: list of classes
        List of available analyser classes for reference
    active_analysis : string or None
        Name of the active analysis
    '''

    def __init__(self):
        
        self.data_directory = []
        self.current_specimen = None
        self.analyser = None
        self.selected_recording = None
        
        self.analyser_class = analyser_classes['motion']
        self.analyser_classes = list(analyser_classes.values())
        
        self.active_analysis = None

        self._folders = {}

        self.groups = SpecimenGroups()

        self.virtual_analysers = []

    
    def create_virtual_analyser(self, name, specimen_names):
    
        analysers = self.get_manalysers(specimen_names)
        virt = VirtualAnalyser(name, analysers)
        self.virtual_analysers.append(virt)

        self.update_gui(changed_specimens=True)

    def set_data_directory(self, data_directory):
        '''
        Update Core's knowledge about the currently selected data_directory.

        Arguments
        ---------
        data_directory : list of strings
            List of paths to the data
        '''
        self.data_directory = data_directory
        
        self._folders = {}
        for data_directory in self.data_directory:
            self._folders[data_directory] = os.listdir(data_directory)


    def set_current_specimen(self, specimen_name):
        '''
        Update Core's knowledge about the currently selected specimen.
        '''
        self.current_specimen = specimen_name
        self.analyser = self.get_manalyser(specimen_name)

    
    def set_selected_recording(self, selected_recording):
        self.selected_recording = selected_recording
    

    def set_analyser_class(self, class_name):
        index = [i for i, ac in enumerate(self.analyser_classes) if ac.__name__ == class_name]
        self.analyser_class = self.analyser_classes[index[0]]
        
        if self.data_directory:
            self.update_gui(changed_specimens=True)


    def list_specimens(self, with_rois=None, with_movements=None, with_correction=None):
        '''
        List specimens in the data directory. May contain bad folders also (no check for contents)

        With the following keyword arguments one select only the specimens fulfilling the conditions
        by setting the keyword argument either to True (has to fulfill condition), False (negative)
        or to None (the condition is not considered)

            with_rois           Specimens with ROIs selected
            with_movements      Specimens with movements measured
            with_correction     Specimens with antenna_level (zero level) correction
        
        For example, if you want the specimens with movements but without antenna_level corrections
        and you won't care about ROIs, set
            with_rois=None, with_movements=True, with_correction=False

        '''
        specimens = []
        
        for data_directory in self.data_directory:
            specimens.extend( [fn for fn in os.listdir(data_directory) if os.path.isdir(os.path.join(data_directory, fn))] )
        
        if with_rois is not None:
            specimens = [specimen for specimen in specimens if self.get_manalyser(specimen, no_data_load=True).are_rois_selected() == with_rois]
        
        if with_movements is not None:
            specimens = [specimen for specimen in specimens if self.get_manalyser(specimen, no_data_load=True).is_measured() == with_movements]
        
        if with_correction is not None:
            specimens = [specimen for specimen in specimens if self.get_manalyser(specimen, no_data_load=True).get_antenna_level_correction() is not False]

        for virt in self.virtual_analysers:
            specimens.insert(0, virt.name)

        return sorted(specimens)
    

    def get_specimen_fullpath(self, specimen_name=None):
        '''
        Returns the full path of a specimen (datadir + specimen_patch)
        
        Arguments
        ---------
        specimen_namse : string or None
            If None use self.current_specimen
        '''
        if specimen_name is None:
            specimen_name = self.current_specimen

        for directory in self.data_directory:
            if specimen_name in self._folders[directory]:
                return os.path.join(directory, specimen_name)

        raise ValueError("no specimen with name {}".format(specimen_name))


    def _configure_analyser(self, analyser):
        if self.active_analysis:
            analyser.active_analysis = self.active_analysis
        return analyser


    def get_manalyser(self, specimen_name, **kwargs):
        '''
        Gets manalyser for the specimen specified by the given name.
        '''
        # Special Case. Virtual analysers are always of the virtual
        # analyser class, no matter what analyser selected
        virtnames = [virt.name for virt in self.virtual_analysers]
        if specimen_name in virtnames:
            return self.virtual_analysers[virtnames.index(specimen_name)]
        
        # Normal case under, nonvirtual analyser

        for directory in self.data_directory:
            if specimen_name in self._folders[directory]:
                break

        analyser = self.analyser_class(directory, specimen_name, **kwargs)

        return self._configure_analyser(analyser)
    

    def get_manalysers(self, specimen_names, **kwargs):
        '''
        Like get_manalyser but returns a list of analyser objects and also
        checks for specimen groups if a specimen cannot be found.
        '''
        analysers = []
        for name in specimen_names:
            try:
                ans = [self.get_manalyser(name, **kwargs)]
            except FileNotFoundError:
                ans = [self.get_manalyser(n, **kwargs) for n in self.groups.groups.get(name, [])]
                
                # Try again and load
                if ans is []:
                    self.groups.load_groups()
                    ans = [self.get_manalyser(n, **kwargs) for n in self.groups.groups.get(name, [])]
            
            if ans is []:
                raise FileNotFoundError('Cannot find specimen {}'.format(name))
            
            for an in ans:
                analysers.append( self._configure_analyser(an) )

        return analysers


    def adm_subprocess(self, specimens, terminal_args, open_terminal=False):
        '''
        Invokes drosom/terminal.py
        
        Arguments
        ---------
        specimens : list of string
            List of specimen names or 'current'
        terminal_args : string
            Agruments passed to the plotter
        open_terminal : bool
            If true open in a cmd window (on Windows) or lxterm (on Linux)
        '''
        
        # 1) Find python executable and wrap the filename by quation marks if spaces present
        python = sys.executable
        if ' ' in python:
            python = '"' + python + '"'

       
        # 2) Find the full path to the adm Python file in the gonio root
        pyfile = os.path.join(CODE_ROOTDIR, 'drosom/terminal.py')
        
        # Check for spaces in the filename. If there are spaces in the filename,
        # we have to encapsulate the filename by quation marks
        if ' ' in pyfile:
            pyfile = '"' + pyfile + '"'

        # 3) Get the specimen directoires (full paths separated by space)
        if specimens == 'current':
            specimen_names = self.analyser.folder
        else:
            specimen_names = ':'.join(specimens)
            if not ':' in specimen_names:
                specimen_names += ':'
        
        if self.active_analysis not in ['default', '', None]:
            terminal_args += ' --active-analysis '+ self.active_analysis

        arguments = '-D "{}" -S "{}" {}'.format(' '.join(self.data_directory), specimen_names, terminal_args)
        
        # Add the type option matching the current analyser
        for key, value in analyser_classes.items():
            if value is self.analyser_class:
                arguments = f'--type {key} {arguments}'

        command = '{} {} {} &'.format(python, pyfile, arguments)
        
        if open_terminal:
            if platform.system() == 'Linux':
                command = 'lxterm -e ' + command
            elif platform.system() == 'Windows':
                command = 'start /wait ' + command
            else:
                raise OSError('Operating system not supported by gonio?')

        print(command)
        
        subprocess.run(command, shell=True)


    def update_gui(self, changed_specimens=False):
        raise ValueError("GUI should overload update_gui method in Core core.py")




'''
General methods common for both DrosoM and DrosoX.
'''

import os
import json
from os import listdir
from os.path import isdir, join


from gonioanalysis.directories import GONIODIR


def simple_select(list_of_strings):
    '''
    Simple command line user interface for selecting a string
    from a list of many strings.
    
    Returns the string selected.
    '''

    for i_option, option in enumerate(list_of_strings):
        print('{}) {}'.format(i_option+1, option))

    while True:
        sel = input('Type in selection by number: ')

        try:
            sel = int(sel)
        except TypeError:
            print('Please input a number')
            continue

        if 1 <= sel <= len(list_of_strings):
            return list_of_strings[sel-1]




class SpecimenGroups:
    '''
    Managing specimens into groups. 
    '''

    def __init__(self):
        self.groups = {}
        
        self.load_groups()

    def load_groups(self):
        try:
            with open(os.path.join(GONIODIR, 'specimen_groups.txt'), 'r') as fp:
                self.groups = json.load(fp)
        except FileNotFoundError:
            print('No specimen groups')

    def save_groups(self):
        '''
        Save groups but only if groups has some members (does not save an empty dict)
        '''
        if len(self.groups.keys()) >= 0:
            with open(os.path.join(GONIODIR, 'specimen_groups.txt'), 'w') as fp:
                json.dump(self.groups, fp)

      
    def new_group(self, group_name, *specimens):
        self.groups[group_name] = [*specimens]
    

    def get_groups(self):
        '''
        Returns the groups dictionary
        '''
        return self.groups

    def get_specimens(self, group_name):
        return self.groups[group_name]


class DrosoSelect:
    '''
    Selecting a Droso folder based on user input or programmatically.

    Folder has to start with "Droso".

    TODO:
    - add programmatic selection methods of folders
    '''

    def __init__(self, datadir=None):
        '''
        datadir     Where the different droso folder are in
        '''
        if datadir is None:
            self.path = input("Input data directory >> ")
        else:
            self.path = datadir

        folders = [fn for fn in os.listdir(self.path) if isdir(join(self.path, fn))]
        self.folders = [os.path.join(self.path, fn) for fn in folders]
 
        self.groups = SpecimenGroups()
    

    def parse_specimens(self, user_input):
        '''
        Parse user input to get manalyser names.

        Arguments
        ---------
        *user_input : string
            Comma separated list of specimen names or indices of self.folders,
            or a group name.
        
        Raises ValueError if unable to parse.
        '''
        
        # 1) If user supplies a specimen group name
        if user_input in self.groups.get_groups().keys():
            user_input = ','.join(self.groups.get_specimens(user_input))

            sel_keys = [os.path.basename(x) for x in user_input.split(',')]
            selections = [folder for folder in self.folders if os.path.basename(folder) in sel_keys]
            
            if len(selections) == len(sel_keys):
                print('Selecting by group.')
            else:
                print('Group is invalid; Specified specimens do not exist')
         
        else:
            # 2) Otherwise first try if indices to self.folders
            try:
                sel_indices = [int(i) for i in user_input.split(',')]
                selections = [self.filt_folders[i] for i in sel_indices]
            except IndexError:
                print('One of the given numbers goes over limits, try again.')
            
            # 3) Next try if specifying by specimen names
            except ValueError:
                print('Not number values given, trying with base names')
                
                sel_keys = [os.path.basename(x) for x in user_input.split(',')]
                selections = [folder for folder in self.folders if os.path.basename(folder) in sel_keys]
                
                if len(selections) == len(sel_keys):
                    print('Worked.')
                else:
                    print('Did not work, try again.')
        
        try:
            selections
        except:
            ValueError("parse_specimens unable to process {}".format(user_input))

        return selections
        

    def ask_user(self, startswith='', endswith='', contains=''):
        '''
        In terminal, ask user to select a Droso folder and can perform simple
        filtering of folders based on folder name.

        INPUT ARGUMENTS         DESCRIPTION
        startswith              Folder's name has to start with this string
        endswith                Folder's name has to have this string in the end
        contains                Folder's name has to have this string somewhere

        RETURNS                 A list of directories (strings) that the user selected?
        '''
        available_commands = ['new_group', 'list_groups', ]

        # Filtering of folders based on their name: startswith, endswith, and contains
        self.filt_folders = [f for f in self.folders if
                os.path.split(f)[1].startswith(startswith) and os.path.split(f)[1].endswith(endswith) and contains in os.path.split(f)[1]]
        
        folders = self.filt_folders

        print('\nSelect a Droso folder (give either number or drosoname, to select many comma split)')
        for i, folder in enumerate(folders):
            print("  {}) {}".format(i, folder))
        
        print('Type help for additional commands')

        while True:
            user_input = input('>> ')

            # 
            splitted = user_input.split(' ')
            if splitted[0] == 'help':
                print('Following commands are avaible')
                for cmd in available_commands:
                    print('  '+cmd)
            elif splitted[0] == 'new_group':
                self.groups.new_group(*splitted[1:])
            elif splitted[0] == 'list_groups':
                print(self.groups.get_groups())
            else:
                try:
                    selections = self.parse_specimens(user_input)
                    break
                except ValueError:
                    pass

        print('\nSelected {}\n'.format(selections))
        
        self.groups.save_groups()
    
        return selections





'''
Plot the results from spatial 2D calibration where the pinhole/fibre was
moved in the camera's (x, y) coordinates while only the green stimulus
LED was turned on and the fibre was connected to the spectrometer.
'''

import tifffile
import numpy as np
import matplotlib.pyplot as plt

from movemeter import Movemeter
from marker import Marker

def get_xy_coordinates(image_fn, match_image):
    '''
    Takes in an image of the pinhole and returs coordinates
    of the pinhole.
    
    image_fn        Image where we look for the match
    match_image     Matching target
    '''
    
    image = tifffile.imread(image_fn)

    movemeter = Movemeter()
    movemeter.set_data(image, )
    #plt.imshow(np.clip(image, np.min(image), np.percentile(image, 50)))
    #plt.show()



def main():
    image_fn = '/home/joni/smallbrains-nas1/array1/pseudopupil_joni/Spectrometer/DPP_cal_1_ty2/snap_2020-02-21_14.15.08.088000_0.tiff'
    
    fig, ax = plt.subplots()
    marker = Marker(fig, ax, [image_fn], None)
    pinhole = marker.run()    
    
    crop = pinhole[image_fn])
    pinhole_image = image_fn


    coodinates = get_xy_coordinates(image_fn)

    

if __name__ == "__main__":
    main()








from .left_right import left_right_summary, left_right_displacements
from .pdf_summary import pdf_summary
from .displacement import mean_folder_repeats

export_functions = {
        'mean_repeats': mean_folder_repeats
        }

export_docstrings = {
        'mean_repeats': 'Displacement "mean repeats" (folder -> its mean)'
        }




'''
Working in 3D space
'''
import math
from math import sin, cos, tan, radians, pi, acos, atan, sqrt, degrees, atan2

import numpy as np


def where_vertical_between(points_3d, lower=None, upper=None, reverse=False):
    ''''Returns a boolean array based on points' vertical angle.

    Takes in 3D points and returns an 1D True/False array of length points_3d
    
    Arguments
    ---------
    points_3d : sequence
        A sequence of (x,y,z) points
    lower : float
        Lower vertical angle degree in degrees
    upper : float
    reverse : bool
        If True, inverses the returned array (True -> False and vice versa).

    Returns
    -------
    booleans : ndarray
        1D True/False array
    '''
    
    # Calculate each point's vertical angle in degrees
    verticals = np.degrees(np.arcsin(points_3d[:,2]/ np.cos(points_3d[:,0]) ))

    # Check each point's y coordinate; If it is negative, we
    # have to fix its angle
    # FIXME: Do this in numpy for better performance
    for i_point in range(len(points_3d)):
        if points_3d[i_point][1] < 0:
            if verticals[i_point] > 0:
                verticals[i_point] = 180-verticals[i_point]
            else:
                verticals[i_point] = -180-verticals[i_point]

    booleans = np.ones(len(points_3d), dtype=bool)
    if lower is not None:
        booleans = booleans * (verticals > lower)
    if upper is not None:
        booleans = booleans * (verticals < upper)
    
    if reverse:
        booleans = np.invert(booleans)

    return booleans
 




def to_spherical(x,y,z, return_degrees=False):
    '''
    Transform to spherical coordinates (ISO)
    
    return_degrees     If true, return angles in degrees instead of radians

    Returns: r, phi, theta
    '''
    r = sqrt(x**2+y**2+z**2)
    phi = atan2(y, x)
    theta = acos(z/(r))
    
    if return_degrees:
        phi = degrees(phi)
        theta = degrees(theta)

    return r, phi, theta

def normalize(P0, P1, scale=1):
    '''
    Normalize a vector
    
    P0      Vector start point
    P1      Vector end point
    '''
    P0 = np.array(P0)
    P1 = np.array(P1)


    vec = P1 - P0
    
    if len(np.nonzero(vec)[0]) == 0:
        return P0

    vec = vec / np.linalg.norm(vec)
    vec *= scale
    vec += P0

    return vec


#def mean_distance(points):
#    return np.mean(np.linalg.norm(points, axis=1))




def nearest_neighbour(point_A, points_B, max_distance=None):
    '''
    Return the nearest point to the point_A from points_B.

    point_A         1D np.array [x0, y0, z0]
    points_B        2D np.array [ [x1,y1,z1], [z2,y2,z2], ... ]
    '''

    distances = np.linalg.norm(points_B - point_A, axis=1)

    i_shortest = np.argmin(distances)

    if max_distance:
        if distances[i_shortest] > max_distance:
            return False

    return i_shortest


def mean_vector(point, vectors):
    '''
    Average vectors and return a vector at point point.

    '''

    av = np.mean(vectors, axis=0)
    if np.linalg.norm(av) != 0:

        av += np.array(point)
        av = force_to_tplane(point, av)



        for i in range(0,len(vectors)):
            wanted_len = np.linalg.norm(vectors[i])

            if wanted_len != 0:
                break
        av -= np.array(point)
        av = (av / np.linalg.norm(av) * wanted_len)
    else:
        av = np.array([0,0,0])
        pass
    #x,y,z = point

    #return (angle_tag, (x, av[0]), (y, av[1]), (z, av[2]) )
    return av




def rotate_about_x(point, degs):
    '''
    Rotate a point in 3D space along the first axis (x-axis).
    '''
    
    c = cos(radians(degs))
    s = sin(radians(degs))

    Rx = np.array([[1,0,0], [0, c, -s], [0, s, c]])

    return np.dot(Rx, np.array(point))
    


def force_to_tplane(P0, P1, radius=1):
    '''
    Forces a vector (P0-P1) on a tangent plane of a sphere but
    retaining the vector's length

    P0 is the point on the sphere (and the tangent plane)
    P1 is the point off the tangent plane

    Returns P2, point on the tangent plane and the line connecting
    the sphere centre point to the P1.

    Notice DOES NOT RETURN VEC BUT P2 (vec=P2-P0)
    '''
    

    a = radius / (P0[0]*P1[0]+P0[1]*P1[1]+P0[2]*P1[2])
    
    return P1 * a


#def projection_to_tplane(P)


def yaw_correct(point, yaw, horizontal, vertical):
    '''Transform yaw

    Only special cases yaw == 90, yaw==0 or yaw == -90 are implemented.
    Free case scenarios may require extra information about the
    rotation zero positions

    Arguments
    ---------
    yaw : float or int
        Rotation of the specimen along its yaw (ie. longitudal rotation)
    '''
    if yaw in [90, 0, -90]:
        return rotate_along_arbitrary(
                np.array((0,1,0)), np.array(point), radians(yaw))
    
    raise NotImplementedError(f"Yaw {yaw} not supported (only -90,0,90)")


def camera2Fly(horizontal, vertical, radius=1):
    '''
    With the given goniometer positions, calculates camera's position
    in fly's cartesian coordinate system.
    
    Input in degrees

    Arguments
    ---------
    horizontal, vertical : float
        Rotation of the two rotation stages, in degrees
    '''
    #print('Horizontal {}'.format(horizontal))
    #print('Vertical {}'.format(vertical))

    h = radians(horizontal)
    v = radians(vertical)
    
    #x = sqrt( radius**2 * (1 - (cos(h) * sin(v))**2 * (tan(v)**2+1)) )
    
    #y = cos(h) * cos(v) * radius
    
    #z = cos(h) * sin(v) * radius

    y = cos(h)*cos(v)*radius
    z = y * tan(v)
    
    # abs added because finite floating point precision
    x = sqrt(abs(radius**2 - y**2 - z**2))
    
    # Make sure zero becomes zero
    if x < 10**-5:
        x = 0

    # Obtain right sign for x
    looped = int(h / (2*pi))
    if not 0 < (h - 2*pi*looped ) < pi:
        x = -x

    return x, y, z


def camera_rotation(horizontal, vertical, return_degrees=False):
    '''
    Camera's rotation
    '''
    
    if vertical > 90:
        vvertical = 180-vertical
    else:
        vvertical = vertical

    rot = -(sin(radians(horizontal))) * radians(vvertical)
    #rot= -((radians(horizontal))/(pi/2)) * radians(vvertical)
    
    if vertical > 90:
        rot += radians(180)
        rot = -rot
  
    if return_degrees:
        rot = degrees(rot)

    return -rot



def camvec2Fly(imx, imy, horizontal, vertical, radius=1, normalize=False):
    '''
    Returns 3D vector endpoints.

    normalize       If true, return unit length vectors
    '''
    # Create unit vectors in camera coordinates
    x,y,z = camera2Fly(horizontal, vertical, radius=radius)

    uimx = np.array(camera2Fly(horizontal, vertical, radius=radius)) - np.array(camera2Fly(horizontal+1, vertical, radius=radius))
    uimx = uimx / np.linalg.norm(uimx)

    uimy = np.array([0, -sin(radians(vertical)), cos(radians(vertical))])

    vector = uimx*imx + uimy*imy

    if normalize:
        length = np.linalg.norm(vector)
        if length != 0:
            
            if type(normalize) == type(42) or type(normalize) == type(4.2):
                length /= normalize
        
            vector = vector / length


    fx, fy, fz = np.array([x,y,z]) + vector 
    return fx, fy, fz

#
#def findDistance(point1, point2):
#    '''
#    Returns PSEUDO-distance between two points in the rotation stages angles coordinates (horizontal_angle, vertical_angle).
#
#    It's called pseudo-distance because its not likely the real 3D cartesian distance + the real distance would depend
#    on the radius that is in our angles coordinate system omitted.
#
#    In the horizontal/vertical angle system, two points may seem to be far away but reality (3D cartesian coordinates)
#    the points are closed to each other. For example, consider points
#        (90, 10) and (90, 70)
#    These points are separated by 60 degrees in the vertical (pitch) angle, but because the horizontal angle is 90 degrees
#    in both cases, they are actually the same point in reality (with different camera rotation)
#
#    INPUT ARGUMENTS     DESCRIPTION
#    point1              (horizontal, vertical)
#
#
#    TODO:   - Implement precise distance calculation in 3D coordinates
#            - ASSURE THAT THIS ACTUALLY WORKS???
#    '''
#    # Scaler: When the horizontal angle of both points is close to 90 or -90 degrees, distance
#    # should be very small
#    scaler = abs(math.sin((point1[0] + point2[0])/ 2))
#    # All this is probably wrong, right way to do this is calculate distances on a sphere
#    return scaler * math.sqrt( (point1[0]-point2[0])**2 + (point1[1]-point2[1])**2 )
#
#
#def findClosest(point1, points, distance_function=None):
#    '''
#    Using findDistance, find closest point to point1.
#    '''
#    
#    distances = []
#
#    if not callable(distance_function):
#        distance_function = findDistance
#
#    for point2 in points:
#        distances.append( distance_function(point1, point2) )
#
#    argmax_i = distances.index(min(distances))
#    
#    return points[argmax_i]
#


def get_rotation_matrix(axis, rot):
    '''
    Returns an elementar rotation matrix.

    axis        'x', 'y', or 'z'
    rot         rotation in radians
    '''
    # Calculate sin and cos terms beforehand
    c = cos(rot)
    s = sin(rot)

    if axis == 'x':
        return np.array([[1,0,0], [0,c,-s], [0,s,c]])
    elif axis == 'y':
        return np.array([[c,0,s], [0,1,0], [-s,0,c]]) 
    elif axis == 'z':
        return np.array([[c,-s,0], [s,c,0], [0,0,1]])
    else:
        raise ValueError('Axis has to be x, y, or z, not {}'.format(axis))


def rotate_along_arbitrary(P1, points, rot):
    '''
    Rotate along arbitrary axis.

    P0 is at origin.
    
    P0 and P1 specify the rotation axis

    Implemented from here:
    http://paulbourke.net/geometry/rotate/
    
    Arguments
    ---------
    P1 : np.ndarray
    points : np.ndarray
    rot : float or int
        Rotation in radians (not degres!)
    '''

    a,b,c = P1 / np.linalg.norm(P1)
    d = math.sqrt(b**2 + c**2)
    
    if d == 0:
        Rx = np.eye(3)
        Rxr = np.eye(3)
    else:
        Rx = np.array([[1,0,0],[0,c/d, -b/d],[0,b/d, c/d]])
        Rxr = np.array([[1,0,0],[0,c/d, b/d],[0,-b/d, c/d]])

    Ry = np.array([[d,0,-a],[0,1,0], [a,0,d]])
    Ryr = np.array([[d,0,a],[0,1,0], [-a,0,d]])

    Rz = get_rotation_matrix('z', rot)

    return (Rxr @ Ryr @ Rz @ Ry @ Rx @ points.T).T


def rotate_points(points, yaw, pitch, roll):
    '''
    Just as rotate_vectors but only for points.
    
    Arguments
    ---------
    yaw, pitch, roll : float or int
        Rotations in radians (not degrees!)
    '''
    yaw_ax = (0,0,1)
    pitch_ax = (1,0,0)
    roll_ax = (0,1,0)

    axes = np.array([yaw_ax, pitch_ax, roll_ax])

    rotations = [yaw, pitch, roll]
    
    for i in range(3):
        points = rotate_along_arbitrary(axes[i], points, rotations[i])
    
    return points



def rotate_vectors(points, vectors, yaw, pitch, roll):
    '''
    In the beginning,  it is assumed that
        yaw     rotation along Z
        pitch   rotation along X
        roll    rotation along Y

    ie that the fly head is at zero rotation, antenna roots pointing towards
    positive y-axis.
    '''

    yaw_ax = (0,0,1)
    pitch_ax = (1,0,0)
    roll_ax = (0,1,0)

    axes = np.array([yaw_ax, pitch_ax, roll_ax])

    rotations = [yaw, pitch, roll]
    
    for i in range(3):
        new_points = rotate_along_arbitrary(axes[i], points, rotations[i])
        new_vectors = rotate_along_arbitrary(axes[i], points+vectors, rotations[i]) - new_points
        
        points = new_points
        vectors = new_vectors


        # Update axes
        #axes = rotate_along_arbitrary(axes[i], axes, rotations[i])
    
    return points, vectors


def distance(a, b):
    '''
    Calculates distance between two points in 3D cartesian space.
    a,b      (x,y,z)
    '''

    return sqrt((a[0]-b[0])**2 + (a[1]-b[1])**2 + (a[2]-b[2])**2)


def optimal_sampling(horizontals, verticals):
    '''
    Determine optimal way to sample using two orthogonal goniometers.
    '''
    
    steps = ((horizontals[1]-horizontals[0]), (verticals[1]-verticals[0]))

    min_distance = 0.75 * distance(camera2Fly(steps[0], steps[1]), camera2Fly(0,0))

  
    goniometer_vals = {}
    
    points = []

    for vertical in verticals:
        goniometer_vals[vertical] = []
        for horizontal in horizontals:
            point = camera2Fly(horizontal, vertical)
            
            append = True
            
            for previous_point in points:
                if distance(previous_point, point) < min_distance:
                    append = False
                    break

            if append:
                points.append(point)
                goniometer_vals[vertical].append(horizontal)

    #for hor, vers in sorted(goniometer_vals.items(), key=lambda x: int(x[0])):
    #    print('{}: {}'.format(hor, vers))
    
    return np.array(points)




def test_rotate_vectors():
    '''

    '''
    
    points = []
    vectors = []
    
    horizontals = np.linspace(-60, 60, 10)
    verticals = np.linspace(0, 180, 10)

    for horizontal in horizontals:
        for vertical in verticals:
            point = camera2Fly(horizontal, vertical)
            vector = np.array([0,0,.2])
            
            P2 = force_to_tplane(point, point+vector)
            
            points.append(point)
            vectors.append(P2-point)
    
    points = np.array(points)
    vectors = np.array(vectors)

    points, vectors = rotate_vectors(points, vectors, radians(0), radians(89), 0)
    
    from gonioanalysis.drosom.plotting import vector_plot



    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')    
    ax.view_init(elev=90, azim=0)
    vector_plot(ax, points, vectors)
    
    plt.savefig('test_rotate_vectors.png')


def test_imx():
    
    P0 = (2, -6)
    
    b = atan(P0[1] / P0[0])
    P1 = (P0[0]-sin(b), P0[1]+cos(b))


    plt.scatter(*P0, color='blue')
    plt.scatter(*P1, color='red')
    plt.scatter(0,0)

    plt.show()


def test_camvec2Fly():
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')    
    ax.scatter(0,0,0, s=10) 
    
    horizontals = np.linspace(0, 360)

    for vertical in horizontals:

        imx, imy, horizontal = (1,0,0) 

        point0 = camera2Fly(horizontal, vertical)
        point1 = camvec2Fly(imx,imy,horizontal,vertical)

        print(point0)
        ax.scatter(*point0, s=10, color='blue')
        ax.scatter(*point1, s=10, color='red')
        



    plt.show()

def test1_camera_rotation():

    from drosom import get_data
    import tifffile
    import matplotlib.pyplot as plt
    from scipy import ndimage
    from gonioimsoft.anglepairs import strToDegrees

    data = get_data('/home/joni/smallbrains-nas1/array1/pseudopupil_imaging/DrosoM23')

    for angle, image_fns in data.items():
        
        horizontal, vertical = strToDegrees(angle)
        rot = degrees(camera_rotation(horizontal, vertical))
        
        im = tifffile.imread(image_fns[0][0])
        im = ndimage.rotate(im, rot)
        
        print('{} {}'.format(horizontal, vertical))
        print(rot)

        plt.imshow(im)
        plt.show()


def test2_camera_rotation():

    horizontals = np.linspace(0,360)
    vertical = 0
    for horizontal in horizontals:
        rot = camera_rotation(horizontal, vertical)
        if rot != 0:
            raise ValueError('rot should be 0 for ALL horizontals when vertical = 0')
        
    horizontals = np.linspace(0,360)
    vertical = 180
    for horizontal in horizontals:
        rot = camera_rotation(horizontal, vertical)
        if round(degrees(rot)) != 180:
            raise ValueError('rot should be 180deg for ALL horizontals when vertical = -180. rot{}'.format(degrees(rot)))
    
    rot = camera_rotation(0, 95)
    if round(degrees(rot)) != 180:
        raise ValueError('rot should be 180deg for 0 horizontal when vertical = -95. rot{}'.format(degrees(rot)))
    
    rot = camera_rotation(-90, 45)
    if round(degrees(rot)) != 45:
        raise ValueError('rot should be 45deg for -90 horizontal when vertical = 45. rot{}'.format(degrees(rot)))
    


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D


    
    #test_camvec2Fly()
    #test_force_to_plane()
    #test1_camera_rotation()
    test_rotate_vectors()




import os

import tkinter as tk
import tkinter.filedialog as filedialog
from tk_steroids.elements import DropdownList

from .common import ImagefolderMultisel

from gonioanalysis.drosom.reports.left_right import left_right_displacements, lrfiles_summarise
from gonioanalysis.drosom.reports.stats import response_magnitudes
import gonioanalysis.drosom.reports as reports


class ExportWizard(tk.Frame):
    '''Export data from selected specimens and image folders.
    '''
    def __init__(self, tk_parent, core):
        tk.Frame.__init__(self, tk_parent)
        self.tk_parent = tk_parent
        self.core = core
        
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.columnconfigure(2, weight=1)
        self.rowconfigure(3, weight=1)

        self.multisel = ImagefolderMultisel(self, self.core, self.on_export_click)
        self.multisel.grid(row=3, column=0, columnspan=3, sticky='NSWE')
        
        self._export_types = reports.export_docstrings.copy()

        # Export type
        tk.Label(self, text='Export type').grid(row=1, column=0)
        self.export_selection = DropdownList(
                self,
                list(self._export_types.keys()),
                fancynames=list(self._export_types.values()))
        self.export_selection.grid(row=2, column=0, sticky='WE')

        # Save location and name
        tk.Label(self, text='Save folder').grid(row=1, column=1)
        self._folder = str(os.getcwd())
        self.folder = tk.Button(self, text=self._folder, command=self.set_folder)
        self.folder.grid(row=2, column=1, sticky='WE')
        
        #self._name = 'group_name'
        tk.Label(self, text='File name').grid(row=1, column=2)
        self.name = tk.Entry(self)
        self.name.grid(row=2, column=2, sticky='WE')


    def set_folder(self):
        directory = filedialog.askdirectory(
                parent=self.tk_parent,
                title='Select save folder for the export',
                mustexist=True,
                initialdir=self._folder)

        if directory:
            self._folder = directory
            self.folder.config(text=directory)

    
    def on_export_click(self, wanted_imagefolders):
        sel = self.export_selection.ticked[0]

        analysers = self.core.get_manalysers(list(wanted_imagefolders.keys()))

        folder = self._folder
        group_name = str(self.name.get())
        
        if sel in reports.export_functions: 
            reports.export_functions[sel](
                    analysers, group_name,
                    wanted_imagefolders=wanted_imagefolders,
                    savedir=folder
                    )
        elif sel == 'Displacement probability TIFF':
            specimens = [';'.join([specimen, *image_folders]) for specimen, image_folders in wanted_imagefolders.items()]
            self.core.adm_subprocess(specimens, '-A magnitude_probability')
        elif sel == 'XY trajectory plot':
            specimens = [';'.join([specimen, *image_folders]) for specimen, image_folders in wanted_imagefolders.items()]
            self.core.adm_subprocess(specimens, '-A xy_trajectory')
        else:
            raise ValueError('Invalid export type selection')

        # Destroy the popup window 
        self.tk_parent.destroy()



'''Main MAnalyser classes for goniometric motion analysis.
'''

import os
import json
import ast
import math
import datetime

import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial import cKDTree as KDTree

from gonioanalysis.drosom.loading import load_data, angles_from_fn, arange_fns
from gonioanalysis.coordinates import (
        camera2Fly,
        camvec2Fly,
        yaw_correct,
        rotate_about_x,
        nearest_neighbour,
        mean_vector,
        optimal_sampling,
        where_vertical_between,
        )
from gonioanalysis.directories import ANALYSES_SAVEDIR, PROCESSING_TEMPDIR
from gonioanalysis.rotary_encoders import to_degrees, step2degree, DEFAULT_STEPS_PER_REVOLUTION

from roimarker import Marker
from movemeter import Movemeter



class AnalyserBase:
    '''Common base class for all analysers.
    
    Features
    --------
    - setting vertical and horizontal ange limits
    - exposing selected attributes as UI options

    Attributes
    ----------
    ui_options : dict
        Info about attributes should be exposed in a user interface. Each item
        is another dictionary with keys "help" (help string) and "type" (type
        conversion function such as int).
    va_limits, ha_limits : list
        Vertical angle and horizontal angle limits for get_3d_vectors.
    '''

    def __init__(self):
        self.ui_options = {}
        
        self.va_limits = [None, None]
        self.ha_limits = [None, None]
        self.alimits_reverse = False


    def get_ui_options(self):
        '''Lists UI options for the analyser: Their values, help and convert.
        '''
        dictionary = []
        for key in self.ui_options:
            dictionary[key] = {}
            dictionary[key]['value'] = getattr(self, key, None)
            dictionary[key]['help'] = self.ui_options[key]['help']
            dictionary[key]['type'] = self.ui_options[key]['type']
        
        return dictionary

    def set_ui_options(self, dictionary):
        '''
        '''
        for key, item in dictionary.items():
            if key in self.ui_options:
                convert = self.ui_options[key].get('type', str)
                setattr(self, key, convert(item))
            else:
                valid_keys = list(self.ui_options.keys())
                raise KeyError(f'Key "{key}" not valid. Valid keys are {valid_keys}')
            

    def set_angle_limits(self, va_limits=(None, None), reverse=False):
        '''
        Limit get_3d_vectors

        All units in degrees.
        '''
        self.va_limits = va_limits
        self.alimits_reverse = reverse

    def __fileOpen(self, fn):
        return self.file_load(fn)

    def __fileSave(self, fn, data):
        return self.file_save(fn, data)

    def file_load(self, fn):
        with open(fn, 'r') as fp:
            data = json.load(fp)
        return data
    
    def file_save(self, fn, data):
        with open(fn, 'w') as fp:
            json.dump(data, fp)
    
    def get_imagefolder(self, image_fn):
        '''Gets the image containing folder (for example, /a/b/c/image -> c).

        Arguments
        ---------
        image_fn : string
            Full path to the image.
        '''
        return os.path.split(os.path.dirname(image_fn))[1]
 

    def get_rotstep_size(self):
        '''
        Returns how many degrees one rotation encoder step was
        (the return value * steps == rotation in degrees)
        '''
        return 360/DEFAULT_STEPS_PER_REVOLUTION

    
    def _merge_by_distance(self, points, vectors, merge_distance):
        '''3D vectors merge by distance
        '''
        # FIXME: Suboptimal performance in this implementation)

        i_point = -1
        while True:
            i_point += 1
            N = len(points)
            
            if i_point >= N:
                break
            
            # Calculate distance of i_point to other points (and
            # itself)
            dists = np.linalg.norm(points[i_point]-points, axis=1)
            
            # Look points that fill the merge distance
            imerge = [i for i in range(N) if (
                dists[i]<merge_distance)]
            
            if not imerge:
                continue
            
            # Calculate merged at i_point
            points[i_point] = np.mean(
                    [points[i] for i in imerge], axis=0)
            vectors[i_point] = np.mean(
                    [vectors[i] for i in imerge], axis=0)

            # Remove merged
            points = np.array([points[i] for i in range(N) if (
                i not in imerge or i==i_point)])
            vectors = np.array([vectors[i] for i in range(N) if (
                i not in imerge or i==i_point)])
    
        return points, vectors


class MAnalyser(AnalyserBase):
    '''Motion analysis for GonioImsoft data.

    Attributes
    ----------
    data_path : string
        A full file path to the specimen folder's location.
    folder : string
        The name of the specimen folder.
    active_analysis : string
        Name of the active analysis
    ROIs
    movements : dict
        Nested dictionary of the measured 2D movements from Movemeter.
        
        Nested structure
        -----------------
        self.movements[eye][angle][i_repeat][x/y/time]
            eye = "left" or "right"
            angle = recording_name.lstrip('pos') // for example angle="(0, 0)_uv"
    eyes : tuple of strings
        Default ("left", "right").
    vector_rotation : float or None
        Rotation of 2D vectors (affects 3D)
    imagefolder_skiplist : dict
    '''

    def __init__(self, data_path, folder, clean_tmp=False, no_data_load=False,
                 active_analysis=''):
        '''Initialize the MAnalyser object.

        Arguments
        ---------
        no_data_load : bool
            If True, skips loading data at constructing the object (use if
            needing many short lived objects.
        active_analysis : string
            Name of the activated analysis
        '''
        super().__init__()
        self._no_data_load = no_data_load
        
        self.ROIs = None 
        self.data_path = data_path
        self.folder = folder

        
        # Skip image_folders. i_repeat
        self.imagefolder_skiplist = {}

        # Python dictionary for linked data
        self.linked_data = {}
        
        self.manalysers = [self]
        self.eyes = ("left", "right")
        self.vector_rotation = None

        # Default yaw==0 when the specimen faces towards the
        # camera (its left and right sides are equally exposed to the cam)
        self.attributes = {
                'yaw': 0,
                }
        
        # Different file or folder paths
        self._rois_skelefn = 'rois_{}{}.json' # specimen_name active_analysis
        self._movements_skelefn = 'movements_{}_{}{}.json' # specimen_name, eye, active_analysis
        self._skiplist_savefn = os.path.join(PROCESSING_TEMPDIR, 'MAnalyser_data', folder, 'imagefolder_skiplist.json')
        self._crops_savefn = os.path.join(PROCESSING_TEMPDIR, 'MAnalyser_data', folder, self._rois_skelefn.format(folder, ''))
        self._movements_savefn = os.path.join(PROCESSING_TEMPDIR, 'MAnalyser_data', folder, self._movements_skelefn.format(folder, '{}', ''))
        self._link_savedir= os.path.join(PROCESSING_TEMPDIR, 'MAnalyser_data', folder, 'linked_data')

        # Save attributes like antenna level correction, yaw
        self._attributes_savefn = os.path.join(PROCESSING_TEMPDIR, 'MAnalyser_data', folder, 'attributes.json')

        
        # Load some things
        if no_data_load == False:
            self.stacks = load_data(os.path.join(self.data_path, self.folder))

            if os.path.isfile(self._skiplist_savefn):
                with open(self._skiplist_savefn, 'r') as fp:
                    self.imagefolder_skiplist = json.load(fp)
            
            if os.path.isfile(self._attributes_savefn):
                attributes = self.file_load(self._attributes_savefn)
                self.attributes = {**self.attributes, **attributes}

            self.antenna_level_correction = self._getAntennaLevelCorrection(folder)

            self.load_linked_data()

            # Ensure the directories where the crops and movements are saved exist
            os.makedirs(os.path.dirname(self._crops_savefn), exist_ok=True)
            os.makedirs(os.path.dirname(self._movements_savefn), exist_ok=True)


        # Set the active analysis, and loads ROIs and movements if available
        self.active_analysis = active_analysis

        self.stop_now = False
    
        # No data load to affect only this constructor
        self._no_data_load = False
        
        # Info data about available UI options
        self.ui_options = {
                'vector_rotation': {'help': 'Vector rotation in the 2D imaging plane', 'type': float}
                }
    

    @property
    def name(self):
        return self.folder
    
    @name.setter
    def name(self, name):
        self.folder=name


    @property
    def active_analysis(self):
        if self.__active_analysis == '':
            return 'default'
        else:
            return self.__active_analysis


    @active_analysis.setter
    def active_analysis(self, name):
        '''Sets the active analysis and loads ROIs an movements.

        Arguments
        ---------
        name : string
            The analysis default is "default" or "" (empty string).
        '''
        
        if name == 'default':
            name = ''

        self.__active_analysis = name
        
        if name == '':
            self._crops_savefn = os.path.join(PROCESSING_TEMPDIR, 'MAnalyser_data', self.folder, self._rois_skelefn.format(self.folder, ''))
            self._movements_savefn = os.path.join(PROCESSING_TEMPDIR, 'MAnalyser_data', self.folder, self._movements_skelefn.format(self.folder, '{}', ''))
        else:
            self._crops_savefn = os.path.join(PROCESSING_TEMPDIR, 'MAnalyser_data', self.folder, self._rois_skelefn.format(self.folder, '_'+name))
            self._movements_savefn = os.path.join(PROCESSING_TEMPDIR, 'MAnalyser_data', self.folder, self._movements_skelefn.format(self.folder, '{}', '_'+name))
    
        if self._no_data_load == False:

            if self.are_rois_selected():
                self.load_ROIs()
            else:
                try:
                    del self.ROIs
                except AttributeError:
                    pass
            if self.is_measured():
                self.load_analysed_movements()
            else:
                try:
                    del self.movements
                except AttributeError:
                    pass
            



    def list_analyses(self):
        '''Returns a list of existing analysis names.
        '''
        
        manalyser_dir = os.path.dirname(self._movements_savefn)

        if os.path.isdir(manalyser_dir):
            fns = [fn for fn in os.listdir(manalyser_dir) if
                    self._movements_skelefn.split('{')[0] in fn and
                    self.eyes[0] in fn]
        else:
            fns = []


        names = []

        for fn in fns:
            # Remove ".json" file ending and split along "_"
            secondlast, last = fn.removesuffix('.json').split('_')[-2:]
            
            if last in self.eyes:
                # Last is the eye (default "noname" analysis)
                names.append('default')
            else:
                # Last is the name of the analysis
                names.append(last)

        return names



    def save_attributes(self): 
        self.file_save(self._attributes_savefn, self.attributes)


    def mark_bad(self, image_folder, i_repeat, i_is_relative=True):
        '''
        Marks image folder and repeat to be bad and excluded
        when loading movements.
        
        Arguments
        ---------
        i_repeat : int or string
            The index of the repeat or "all".
        i_is_relative : bool
            If True, i_repeat is relative to what has been previously
            marked bad. If False, then i_repeat is the actual
            index of the repeat.

            For example: consider repeats [0,1,2,3,4,5]
            Consider 4 has been previously marked bad, then the user
            sees the repeats [0,1,2,3,5].
            The task is to mark 5 bad.
            If i_is_relative==False, then i_repeat=5 to mark 5 bad
            If i_is_relative==True, then i_repeat=4 to mark 5 bad
        '''
        
        if self.imagefolder_skiplist.get(image_folder, None) is None:
            self.imagefolder_skiplist[image_folder] = []
        
        # Skiplist of this image_folder
        skiplist = set(self.imagefolder_skiplist[image_folder])

        print(skiplist)
        if i_is_relative and i_repeat != 'all':
            # Inefficient algo if the repeat count is millions or
            # something crazy but not important
            N = self.get_movements_from_folder(image_folder)
            N = max([len(N.get(eye,[])) for eye in self.eyes]) + len(skiplist)
            nonexcluded = [i for i in range(N) if i not in skiplist]

            original = i_repeat
            i_repeat = nonexcluded[i_repeat]

            print(f'Mark bad: {image_folder} abs-i {i_repeat} (rel-i {original})')
        else:
            print(f'Mark bad: {image_folder} abs-i {i_repeat}')

        skiplist.add(i_repeat)
        self.imagefolder_skiplist[image_folder] = list(skiplist)
         
        # Save the complete skiplist on disk
        with open(self._skiplist_savefn, 'w') as fp:
            json.dump(self.imagefolder_skiplist, fp)



    def list_rotations(self, list_special=True, special_separated=False,
            horizontal_condition=None, vertical_condition=None,
            _return_imagefolders=False):
        '''
        List all the imaged vertical-horizontal pair rotations.
        
        Arguments
        ---------
        list_special : bool
            If false, include only rotations whose folders have no suffix
        special_separated : bool
            If true, return standard and special image folders separetly.
        horizontal_condition : callable or None
            A callable, that when supplied with horizontal (in steps),
            returns either True (includes) or False (excludes).
        vertical_condition : callable or None
            Same as horizontal condition but for vertical rotation

        Returns
        -------
        rotations : list of tuples
            List of rotations.
        '''

        def check_conditions(vertical, horizontal):
            if callable(horizontal_condition):
                if not horizontal_condition(horizontal):
                    return False
            if callable(vertical_condition):
                if not vertical_condition(vertical):
                    return False
            return True
 
        standard = []
        special = []

        for key in self.stacks.keys():
            try:
                horizontal, vertical = ast.literal_eval(key)
                if check_conditions(vertical, horizontal) == False:
                    continue

            except (SyntaxError, ValueError):
                # This is now a special folder, ie. with a suffix or something else
                # Try to get the angle anyhow
                splitted = key.replace('(', ')').split(')')
                if len(splitted) == 3:
                    try:
                        horizontal, vertical = splitted[1].replace(' ', '').split(',')
                        horizontal = int(horizontal)
                        vertical = int(vertical)
                        
                        if check_conditions(vertical, horizontal) == False:
                            continue
                    except:
                        pass
                
                if _return_imagefolders:
                    special.append('pos'+key)
                else:
                    special.append((horizontal, vertical))
                continue
            except TypeError:
                # Fully custom folder name
                special.append(key)
                continue

            if _return_imagefolders:
                standard.append('pos'+key)
            else:
                standard.append((horizontal, vertical))
        
        if not list_special:
            special = []

        if special_separated:
            return standard, special
        else:
            return standard + special

    
    def list_imagefolders(self, endswith='', only_measured=False, **kwargs):
        '''
        Returns a list of the image folders (specimen subfolders that contain
        the images).
        
        Arguments
        ---------
        only_measured : bool
            Return only image_folders with completed movement analysis
        
        See list_rotations for other allowed keyword arguments. 
        
        Returns
        -------
        image_folders : list of strings
        '''

        image_folders, special_image_folders = self.list_rotations(
                special_separated=True,
                _return_imagefolders=True,
                **kwargs)
        
        all_folders = [fn for fn in sorted(image_folders) + sorted(special_image_folders) if fn.endswith(endswith)]
        
        if only_measured:
            all_folders = [fn for fn in all_folders if self.folder_has_movements(fn)]

        return all_folders


    def get_horizontal_vertical(self, image_folder, degrees=True):
        '''
        Tries to return the horizontal and vertical for an image folder.

        image_folder
        degrees             If true, return in degrees
        '''
        # Trusting that ( and ) only reserved for the angle
        splitted = key.replace('(', ')').split(')')
        if len(splitted) == 3:
            horizontal, vertical = splitted[1].replace(' ', '').split(',')
            horizontal = int(horizontal)
            vertical = int(vertical)
        
        if degrees:
            return step2degree(horizontal), step2degree(vertical)
        else:
            return horizontal, vertical


    def get_specimen_directory(self):
        return os.path.join(self.data_path, self.folder)

   
    def get_N_repeats(self, image_folder):
        '''Returns how many repeats are there in the image folder
        
        Arguments
        ---------
        image_folder : str
            Name of the image folder with the pos prefix
        '''
        folder = image_folder.removeprefix('pos')
        return len(self.stacks[folder])


    def list_images(self, image_folder, i_repeat=None,
                    absolute_path=False):
        '''List all image filenames in an image folder
        
        FIXME: Alphabetical order not right because no zero padding
        
        Arguments
        ---------
        image_folder : str
            Name of the image folder with the pos prefix
        i_repeat : int or None
            If None, return all images.
        absolute_path : bool
            If true, return filenames with absolute path instead of
            relative

        '''

        fns = []
        
        folder = image_folder.removeprefix('pos')

        if i_repeat is None:
            # Flatten out the i_repeat structure
            for ims in self.stacks[folder]:
                fns.extend(ims)

        elif isinstance(i_repeat, int):
            fns.extend(self.stacks[folder][i_repeat])
        else:
            raise TypeError(
                    f"i_repeat has to be int or None, got {i_repeat}")

        if not absolute_path:
            fns = [os.path.basename(fn) for fn in fns]

        return fns

   
    def get_specimen_name(self):
        '''
        Return the name of the data (droso) folder, such as DrosoM42
        '''              
        return self.folder
   
    
    @staticmethod
    def _getAntennaLevelCorrection(fly_name): 
        fn = os.path.join(ANALYSES_SAVEDIR, 'antenna_levels', fly_name+'.txt')

        if os.path.exists(fn):
            with open(fn, 'r') as fp:
                antenna_level_offset = float(fp.read())
        else:
            antenna_level_offset = False
        
        return antenna_level_offset

    
    def _correctAntennaLevel(self, angles):
        '''
        angles  In degrees, tuples, (horizontal, pitch)
        '''
        if self.antenna_level_correction != False:
            for i in range(len(angles)):
                angles[i][1] -= self.antenna_level_correction

        return angles


    def get_antenna_level_correction(self):
        '''
        Return the antenna level correction or if no correction exists, False.
        '''
        return self._getAntennaLevelCorrection(self.folder)


    def get_imaging_parameters(self, image_folder):
        '''
        Returns a dictionary of the Gonio Imsoft imaging parameters.
        The dictionary is empty if the descriptions file is missing.

        image_folder : string
        '''
        
        parameters = {}


        fn = os.path.join(self.data_path, self.folder, image_folder, 'description.txt')
        
        if not os.path.isfile(fn):
            # Fallback for older Imsoft data where only
            # one descriptions file for each imaging
            old_fn = os.path.join(self.data_path, self.folder, self.folder+'.txt')
            if os.path.isfile(old_fn):
                fn = old_fn
            else:
                return {}


        with open(fn, 'r') as fp:
            for line in fp:
                if line.startswith('#') or line in ['\n', '\r\n']:
                    continue
                split= line.strip('\n\r').split(' ')

                if len(split) >= 1:
                    parameters[split[0]] = split[1]
                else:
                    parameters[split[0]] = ''
        
        return parameters

    def get_specimen_age(self):
        '''
        Returns age of the specimen, or None if unkown.

        If many age entries uses the latest for that specimen.
        '''
        
        try:
            self.descriptions_file
        except AttributeError:
            self.descriptions_file = self._load_descriptions_file()
        
        for line in self.descriptions_file[::-1]:
            if line.startswith('age '):
                return line.lstrip('age ')

        return None


    def get_specimen_sex(self):
        '''
        Returns sex of the specimen, or None if unkown.

        If many sex entries uses the latest for that specimen.
        '''
        
        try:
            self.descriptions_file
        except AttributeError:
            self.descriptions_file = self._load_descriptions_file()
        
        for line in self.descriptions_file[::-1]:
            if line.startswith('sex '):
                return line.lstrip('sex ').strip(' ').strip('\n')

        return None

    
    def get_imaging_frequency(self, image_folder, fallback_value=100.):
        '''
        Return imaging frequency (how many images per second) for an image folder
        by searching for frame_length field in the descriptions file.

        Arguments
        ---------
        image_folder : string or None
            Name. If None (not recommended), returns the imaging frequency from
            any image folder that is available (~random).
        fallback_value : float or None
            What to return if fs cannot be determined.

        Returns
        -------
        fs : float
            The imaging frequency. Default of 100 HzNone if the imaging frequency could not be determined.
        '''
        if image_folder is None:
            folders = self.list_imagefolders()
        else:
            folders = [image_folder]
        
        for folder in folders:
            fs = self.get_imaging_parameters(folder).get('frame_length', None)
            if fs is not None:
                break

        if fs is None:
            # FIXME
            return fallback_value
        else:
            return 1/float(fs)
    
    def get_pixel_size(self, image_folder):
        '''
        Return the pixel size of the imaging.
        Currently always returns the same static value of 1.22375. 
        '''
        # Based on the stage micrometer;
        # 0.8 µm in the images 979 pixels 
        return 1/1.22376
    

    def get_snap_fn(self, i_snap=0, absolute_path=True):
        '''
        Returns the first snap image filename taken (or i_snap'th if specified).

        Many time I took a snap image of the fly at (0,0) horizontal/vertical, so this
        can be used as the "face photo" of the fly. 
        '''

        snapdir = os.path.join(self.data_path, self.folder, 'snaps')
        fns = [fn for fn in os.listdir(snapdir) if fn.endswith('.tiff')]
        fns.sort()

        if absolute_path:
            fns = [os.path.join(snapdir, fn) for fn in fns]

        return fns[i_snap]


    def load_ROIs(self):
        '''
        Load ROIs (selected before) for the left/right eye.
        
        INPUT ARGUMENTS     DESCRIPTION
        eye                 'left' or 'right'
        
        DETAILS
        While selecting ROIs, both eyes are selcted simultaneously. There's
        no explicit information about from which eye each selected ROI is from.
        Here we reconstruct the distinction to left/right using following way:
            1 ROI:      horizontal angle determines
            2 ROIs:     being left/right in the image determines
        
        Notice that this means that when the horizontal angle is zero (fly is facing towards the camera),
        image rotation has to be so that the eyes are on image's left and right halves.
        '''

        self.ROIs = {'left': {}, 'right': {}}

        with open(self._crops_savefn, 'r') as fp:
            marker_markings = json.load(fp)

        for image_fn, ROIs in marker_markings.items():
            
            # Since use to relative filenames in the ROIs savefile
            image_fn = os.path.join(self.data_path, self.folder, image_fn)

            # ROIs smaller than 7 pixels a side are not loaded
            good_rois = []
            for i_roi in range(len(ROIs)):
                if not (ROIs[i_roi][2] < 7 and ROIs[i_roi][3] < 7):
                    good_rois.append(ROIs[i_roi])
            ROIs = good_rois

            pos = self.get_imagefolder(image_fn)
            try:
                horizontal, pitch = angles_from_fn(pos)
            except:
                horizontal, pitch = (0, 0)
           
            pos = pos.removeprefix('pos')

            if '_cam' in os.path.basename(image_fn):
                # Allows cameras 0-9
                try:
                    i_camera = int(image_fn[image_fn.index('_cam') + 4])
                    pos += f'_cam{i_camera}'
                except: pass

            # ROI belonging to the eft/right eye is determined solely by
            # the horizontal angle when only 1 ROI exists for the position
            if len(ROIs) == 1:
                
                if horizontal > 0:
                    self.ROIs['left'][pos] = ROIs[0]
                else:
                    self.ROIs['right'][pos] = ROIs[0]

            # If there's two ROIs
            elif len(ROIs) == 2:
                
                if ROIs[0][0] > ROIs[1][0]:
                    self.ROIs['left'][pos] = ROIs[0]
                    self.ROIs['right'][pos] = ROIs[1]
                else:
                    self.ROIs['left'][pos]= ROIs[1]
                    self.ROIs['right'][pos] = ROIs[0]
            
            elif len(ROIs) > 2:

                # With drift_correction, many ROIs are possible
                #if self.active_analysis == 'drift_correction':
                #    for i_roi, roi in enumerate(ROIs):
                #        self.ROIs[f'roi{i_roi}'] = roi

                print('Warning. len(ROIs) == {} for {}'.format(len(ROIs), image_fn))

        self.N_folders_having_rois = len(marker_markings)
        
        

    def select_ROIs(self, **kwargs):
        '''
        Selecting the ROIs from the loaded images.
        Currently, only the first frame of each recording is shown.

        kwargs      Passed to the marker constructor
        '''
        
        to_cropping = [stacks[0][0] for str_angles, stacks in self.stacks.items()]

        fig, ax = plt.subplots()
        marker = Marker(fig, ax, to_cropping, self._crops_savefn,
                relative_fns_from=os.path.join(self.data_path, self.folder), **kwargs)
        marker.run()


    def are_rois_selected(self):
        '''
        Returns True if a file for crops/ROIs is found.
        '''
        return os.path.exists(self._crops_savefn)


    def count_roi_selected_folders(self):
        '''
        Returns the number of imagefolders that have ROIs selected
        '''
        if self.are_rois_selected():
            return self.N_folders_having_rois
        else:
            return 0


    def folder_has_rois(self, image_folder):
        '''
        Returns True if for specified image_folder at least one
        ROI exsits. Otherwise False.
        ''' 
        try:
            self.ROIs
        except AttributeError:
            return False

        if self.get_rois(image_folder) != []:
            return True

        return False


    def get_rois(self, image_folder):
        rois = []
        for eye in ['left', 'right']:
            try:
                roi = self.ROIs[eye][image_folder.removeprefix('pos')]
                rois.append(roi)
            except:
                continue
        return rois


    def is_measured(self):
        '''
        Returns (True, True) if analyseMovement results can be found for the fly and bot eyes.
        '''
        return all((os.path.exists(self._movements_savefn.format('left')), os.path.exists(self._movements_savefn.format('right'))))



    def folder_has_movements(self, image_folder):
        '''
        Returns True if for specified image_folder has movements
        measured. Otherwise False.
        '''
        try:
            self.movements
        except AttributeError:
            return False

        if any([image_folder.removeprefix('pos') in self.movements[eye].keys()] for eye in ['left', 'right']):
            if len(self.get_displacements_from_folder(image_folder)) > 0:
                return True
        return False


    def load_analysed_movements(self):
        self.movements = {}
        with open(self._movements_savefn.format('right'), 'r') as fp:
            self.movements['right'] = json.load(fp)
        with open(self._movements_savefn.format('left'), 'r') as fp:
            self.movements['left'] = json.load(fp)
        
        if self.__active_analysis == 'drift_correction':
            for key in self.movements['right']:
                if key not in self.movements['left']:
                    self.movements['left'] = self.movements['right']

        # Special analysis name: "drift_correction". If present, subtracted
        # from all the active analysis
        if 'drift_correction' in self.list_analyses() and self.__active_analysis != 'drift_correction':
            dc_analyser = MAnalyser(
                    self.data_path, self.folder, active_analysis='drift_correction')
            
            for eye in ['left', 'right']:
                for image_folder, data in dc_analyser.movements[eye].items():
                    if image_folder not in self.movements[eye]:
                        continue
                    for i_repeat in range(len(data)):
                        A = self.movements[eye][image_folder][i_repeat]
                        B = data[i_repeat]
                        for key in ['x', 'y']:
                            dcx = np.array(list(range(len(B[key]))))
                            fit = np.polynomial.polynomial.Polynomial.fit(dcx, B[key],1)
                            corrected = (np.array(A[key]) - fit(dcx) ).tolist()
                            #corrected = [a-b for a, b in zip(A[key], B[key])]
                            self.movements[eye][image_folder][i_repeat][key] = corrected


        if self.imagefolder_skiplist:
            
            for image_folder, skip_repeats in self.imagefolder_skiplist.items():
                for eye in self.eyes:

                    if self.movements[eye].get(image_folder.removeprefix('pos'), None) is None:
                        continue
                    
                    # Iterate repeats reversed so we can just pop things
                    for i_repeat in sorted(skip_repeats)[::-1]:
                        self.movements[eye][image_folder.removeprefix('pos')].pop(i_repeat)

                    


    def measure_both_eyes(self, **kwargs):
        '''
        Wrapper to self.measure_movement() for both left and right eyes.
        '''
        for eye in self.ROIs:
            self.measure_movement(eye, **kwargs)


    def measure_movement(self, eye, only_folders=None,
            max_movement=30, absolute_coordinates=False, join_repeats=False,
            stop_event=None):
        '''
        Performs cross-correlation analysis for the selected ROIs (regions of interest)
        using Movemeter package.

        If ROIs haven't been selected, calls method self.selectROIs.
        Movements are saved into a tmp directory.

        INPUT ARGUMENTS         DESCRIPTION
        eye                     'left' or 'right'
        only_folders            Analyse only image folders in the given list (that is only_folders).
        max_movement            Maximum total displacement in x or y expected. Lower values faster.
        absolute_coordinates    Return movement values in absolute image coordinates
        join_repeats            Join repeats together as if they were one long recording.
        stop_event              None or threading.Event for stopping the movement measurement
            

        Cross-correlation analysis is the slowest part of the DrosoM pipeline.
        '''
        
        self.movements = {}
        
        if not os.path.exists(self._crops_savefn):
            self.selectROIs() 
        self.load_ROIs()
        

        angles = []
        stacks = []
        ROIs = []

        if not self.ROIs[eye] == {}:

            for angle in self.stacks:
                #if angle in str(self.ROIs[eye].keys()):
                
                # Continue if no ROI for this eye exists
                try :
                    self.ROIs[eye][angle]
                except KeyError:
                    continue

                # Continue if only_folders set and the angle is not in
                # the only folders
                if only_folders and not 'pos'+angle in only_folders:
                    continue

                if join_repeats:
                    fused = []
                    for i_repetition in range(len(self.stacks[angle])):
                        fused += self.stacks[angle][i_repetition]
                
                    self.stacks[angle] = [fused]

                for i_repetition in range(len(self.stacks[angle])):
                    angles.append(angle)
                    stacks.append( self.stacks[angle][i_repetition] )
                    ROIs.append( [self.ROIs[eye][angle]] )

            
            if ROIs == []:
                return None


            # Old upscale was 4
            meter = Movemeter(upscale=10, absolute_results=absolute_coordinates)
            meter.set_data(stacks, ROIs)
            
            for stack_i, angle in enumerate(angles):
                
                if stop_event and stop_event.is_set():
                    self.stop_now = True

                if self.stop_now:
                    self.stop_now = False
                    self.movements = {}
                    print('{} EYE CANCELLED'.format(eye.upper()))
                    return None

                
                print('Analysing {} eye, motion from position {}, done {}/{} for this eye'.format(eye.upper(), angle, stack_i+1, len(ROIs)))

                print("Calculating ROI's movement...")
                x, y = meter.measure_movement(stack_i, max_movement=max_movement)[0]
                
                print('Done.')
                
                try:
                    self.movements[angle]
                except KeyError:
                    self.movements[angle] = []
                
                tags = meter.get_metadata(stack_i)['Image ImageDescription'].values.split('"')
                
                # GonioImsoft start time tag in the images
                if 'start_time' in tags:
                    time = tags[tags.index('start_time') + 2]
                else:
                    time = None

                self.movements[angle].append({'x': x, 'y':y, 'time': time})

        else:
            self.movements = {}
            
        # If only_folders set ie. only some angles were (re)measured,
        # load previous movements also for saving
        if only_folders:
            with open(self._movements_savefn.format(eye), 'r') as fp:
                 previous_movements = json.load(fp)
            
            # Update previous movements with the new movements and set
            # the updated previous movements to be the current movements
            previous_movements.update(self.movements)
            self.movements = previous_movements


        # Save movements
        with open(self._movements_savefn.format(eye), 'w') as fp:
            json.dump(self.movements, fp)
        
        
        #for key, data in self.movements.items():
        #    plt.plot(data['x'])
        #    plt.plot(data['y'])
        #    plt.show()
    


    def get_time_ordered(self, angles_in_degrees=True, first_frame_only=False,
            exclude_imagefolders=[]):
        '''
        Get images, ROIs and angles, ordered in recording time for movie making.
        
        exclude_imagefolders : list
            Imagefolders to exclude

        Returns 3 lists: image_fns, ROIs, angles
                image_fns
        '''
        self.load_ROIs()

        times_and_data = []
        seen_angles = []

        for eye in self.movements:
            for angle in self.movements[eye]:
                
                if 'pos'+angle in exclude_imagefolders:
                    continue

                if not angle in seen_angles: 
                    time = self.movements[eye][angle][0]['time']
                    
                    fn = self.stacks[angle][0]
                    ROI = self.get_moving_ROIs(eye, angle)
                    deg_angle = [list(ast.literal_eval(angle.split(')')[0]+')' ))]
                    
                    if angles_in_degrees:
                        to_degrees(deg_angle)
                    
                    deg_angle = [deg_angle[0] for i in range(len(fn))]

                    times_and_data.append([time, fn, ROI, deg_angle])
                    seen_angles.append(angle)
        
        # Everything gets sorted according to the time
        times_and_data.sort(key=lambda x: x[0])
        
        image_fns = []
        ROIs = []
        angles = []
        
        if not first_frame_only:
            for time, fns, ROI, angle in times_and_data:
                image_fns.extend(fns)
                ROIs.extend(ROI)
                angles.extend(angle)
        else:
            for time, fns, ROI, angle in times_and_data:
                image_fns.append(fns[0])
                ROIs.append(ROI[0])
                angles.append(angle[0])
        
        return image_fns, ROIs, angles


    def get_movements_from_folder(self, image_folder):
        '''
        
        '''
        data = {}
        for eye in ['left', 'right']:
            try:
                data[eye] = self.movements[eye][image_folder.removeprefix('pos')]
            except KeyError:
                pass
        
        return data


    def get_displacements_from_folder(self, image_folder):
        '''
        Returns a list of 1D numpy arrays, which give displacement
        over time for each repeat.

        If no displacement data, returns an empty list

        Calculated from separete (x,y) data
        '''
        displacements = []
        
        for eye, data in self.get_movements_from_folder(image_folder).items():
            for repetition_data in data:
                x = repetition_data['x']
                y = repetition_data['y']
                mag = np.sqrt(np.asarray(x)**2 + np.asarray(y)**2)
                displacements.append(mag)

        return displacements


    def get_raw_xy_traces(self, eye):
        '''
        Return angles, values
        angles      Each recorded fly orientation in steps
        values      X and Y
        '''
        angles = [list(ast.literal_eval(angle)) for angle in self.movements[eye]]
        movement_dict = [self.movements[eye][str(angle)] for angle in angles]
        
        return angles, movement_dict
    

    def get_2d_vectors(
            self, eye, image_folder=None,
            mirror_horizontal=True, mirror_pitch=True,
            correct_level=True, repeats_separately=False,
            mirror_movements=False):
        '''Creates 2D vectors from the movements analysis data.
        
        Arguments
        ---------
        eye : string
            "left" or "right"
        image_folder : None or string
        mirror_horizontal : bool
        mirror_pitch : bool
            Should make so that the negative values are towards
            dorsal and positive towards frontal
            (this is how things on DrosoX were)
        correct_level : bool
        repeats_separately : bool
        mirror_movements : bool
            If True, mirrors the movement directions (X=-X and Y=-Y)

        Returns
        -------
        angles : list
        X, Y : list
            Vector start point: ROI's position at the first frame
            Vector end point: ROI's position at the last frame
        '''

        if image_folder is None:
            # Make the order of angles deterministic
            sorted_angle_keys = sorted(self.movements[eye])
        else:
            key = image_folder.removeprefix('pos')
            if key in self.movements[eye]:
                sorted_angle_keys = [key]
            else:
                sorted_angle_keys = []
        
        angles = [list(ast.literal_eval(angle.split(')')[0]+')' )) for angle in sorted_angle_keys]
        values = [self.movements[eye][angle] for angle in sorted_angle_keys]    

        to_degrees(angles)
        
        if correct_level:
            angles = self._correctAntennaLevel(angles)
        
        if mirror_horizontal:
            for i in range(len(angles)):
                angles[i][0] *= -1
            xdirchange = -1
        else:
            xdirchange = 1
        
        if mirror_pitch:
            for i in range(len(angles)):
                angles[i][1] *= -1

        # Vector X and Y components
        # Fix here if repetitions are needed to be averaged
        # (don't take only x[0] but average)
        if repeats_separately:
            tmp_angles = []            
            X = []
            Y = []
            
            for angle, val in zip(angles, values):
                for repeat in val:
                    tmp_angles.append(angle)
                    X.append( xdirchange*(repeat['x'][-1]-repeat['x'][0]) )
                    Y.append( repeat['y'][-1]-repeat['y'][0] )

            angles = tmp_angles

        else:

            X = [xdirchange*(x[0]['x'][-1]-x[0]['x'][0]) for x in values]
            Y = [x[0]['y'][-1]-x[0]['y'][0] for x in values]
        
        if mirror_movements:
            X = [-x for x in X]
            Y = [-y for y in Y]

        if self.vector_rotation:
            r = math.radians(self.vector_rotation)
            for i in range(len(X)):
                x = X[i]
                y = Y[i]
                
                if angles[i][1] > 0:
                    sr = -1 * r
                else:
                    sr = 1 * r

                if eye == 'left':
                    sr = sr
                elif eye == 'right':
                    sr = -sr

                X[i] = x * math.cos(sr) - y * math.sin(sr)
                Y[i] = x * math.sin(sr) + y * math.cos(sr)

        return angles, X, Y


            
    def get_magnitude_traces(self, eye, image_folder=None,
            mean_repeats=False, mean_imagefolders=False,
            microns=False, _phase=False, _derivative=False):
        '''
        Get all movement magnitudes (sqrt(x**2+y**2)) from the specified eye.
        The results are returned as a dictionary where the keys are the
        angle pairs (self.movements keys)

        eye : string or None
            "left" or "right".
            None leads to taking mean where eyes overlap
        image_folder : string
            If specified, return movements from this image folder.
            Otherwise by default None, movements from all image folders.
        mean_repeats : bool
            Wheter take mean of the mean repeats.
        mean_imagefolders : bool
            Only makes sense when image_folder is None
        mean_eyes : bool
            Only makes sense when eye is None
        microns : bool
            Call self.get_pixel_size(image_folder) to convert from
            pixel units to micrometers.
        _phase: bool
            If true return phase in degrees instead.

        Returns
            if mean_repeats == True
                magnitude_traces = {angle_01: [mag_mean], ...}
            if mean_repeats == False
                magnitude_traces = {angle_01: [mag_rep1, mag_rep2,...], ...}
            
            if mean_imagefolders, there's only one key 'mean'

        '''
        alleye_magnitude_traces = {}
        
        if eye is None:
            eyes = self.eyes
        else:
            eyes = [eye]

        if image_folder is None:
            movement_keys = set().union(*[list(self.movements[eye].keys()) for eye in eyes])
        else:
            movement_keys = [image_folder.removeprefix('pos')]
        
        for eye in eyes:
            magnitude_traces = {}
            for angle in movement_keys:
                
                if self.movements[eye].get(angle, None) is None:
                    # Continue if data for this eye
                    continue

                magnitude_traces[angle] = []
                
                for i_repeat in range(len(self.movements[eye][angle])):
                    x = self.movements[eye][angle][i_repeat]['x']
                    y = self.movements[eye][angle][i_repeat]['y']

                    if _phase:
                        mag = np.degrees(np.arctan2(y, -np.asarray(x)))
                    else:
                        mag = np.sqrt(np.asarray(x)**2 + np.asarray(y)**2)
                    
                    if _derivative:
                        mag = np.diff(mag)

                    magnitude_traces[angle].append( mag )
                
                if mean_repeats:
                    magnitude_traces[angle] = [np.mean(magnitude_traces[angle], axis=0)]

            if magnitude_traces == {}:
                # If nothing for this eye
                continue

            if mean_imagefolders:
                tmp = np.mean([val for val in magnitude_traces.values()], axis=0)
                
                magnitude_traces = {}
                magnitude_traces['imagefoldersmean'] = tmp

            alleye_magnitude_traces[eye] = magnitude_traces

        


        if len(eyes) > 1:
            merge = {}
            # Merge (mean) eyes where one imagefolder hols data from both eyes

            angles = [list(val.keys()) for val in alleye_magnitude_traces.values()]
            angles = set().union(*angles)
            
            for angle in angles:

                data = [alleye_magnitude_traces.get(eye, {}).get(angle, None) for eye in eyes]
                data = [d for d in data if d is not None]

                merge[angle] = np.mean(data, axis=0)


            magnitude_traces = merge

        if microns and not _phase:
            for image_folder in magnitude_traces:
                pixel_size = self.get_pixel_size(image_folder)
                magnitude_traces[image_folder] = [t*pixel_size for t in magnitude_traces[image_folder]]


        return magnitude_traces

    
    def get_moving_ROIs(self, eye, angle, i_repeat=0):
        '''
        Returns a list of ROIs how they move over time.
        Useful for visualizing.
        '''

        moving_ROI = []
        
        if not self.ROIs:
            self.load_ROIs()

        movements = self.movements[eye][angle][i_repeat]
        rx,ry,rw,rh = self.ROIs[eye][angle]
        
        for i in range(len(movements['x'])):
            x = -movements['x'][i]
            y = -movements['y'][i]
            
            moving_ROI.append([rx+x,ry+y,rw,rh])
        return moving_ROI
        
    
    def get_3d_vectors(
            self, eye, image_folder=None,
            return_angles=False, correct_level=True,
            repeats_separately=False, normalize_length=0.1,
            strict=None, vertical_hardborder=None,
            merge_distance=None):
        '''Returns 3D vectors and their starting points.
        
        Arguments
        ---------
        eye : string
            "left" or "right"
        image_folder : string or None
            If not none, return the 3D vectors from this foldted 
        correct_level : bool
            Use estimated antenna levels
        repeats_separately : bool
        normalize_length : float or None
            Normalize vectors to the given length. If None,
            does not normalize. Default 0.1
        strict : None
        vertical_hardborder : None
        merge_distance : None or float
            If not None, merge (take mean) vectors that are closer
            than the given distance.

        Returns
        -------
        points
        vectors
        '''
        
        yaw = self.attributes['yaw']

        # Dirty fix yeat again... When yaw +90 or -90, assume that
        # all vectors are from either the left or the right eye
        # Needed for
        #   - correct L/R coloring when yaw +90, -90
        #   - merge distance to work with 3 part scans
        #
        if yaw == 0:
            # Normal case
            angles, X, Y = self.get_2d_vectors(
                    eye, image_folder=image_folder,
                    mirror_pitch=False, mirror_horizontal=True,
                    correct_level=False, repeats_separately=repeats_separately)
        else:
            if (eye == 'left' and yaw == -90) or (eye == 'right' and yaw == 90):
                return [[],[]]
            
            angles, X, Y = self.get_2d_vectors(
                    'right', image_folder=image_folder,
                    mirror_pitch=False, mirror_horizontal=True,
                    correct_level=False, repeats_separately=repeats_separately)           
            
            angles2, X2, Y2 = self.get_2d_vectors(
                    'left', image_folder=image_folder,
                    mirror_pitch=False, mirror_horizontal=True,
                    correct_level=False, repeats_separately=repeats_separately)
            angles += angles2
            X += X2
            Y += Y2
            #angles = np.concatenate((angles, angles2))
            #X = np.concatenate((X,X2))
            #Y = np.concatenate((Y,Y2))

        N = len(angles)

        points = np.zeros((N,3))
        vectors = np.zeros((N,3))

       
        for i, (angle, x, y) in enumerate(zip(angles, X, Y)):
            horizontal, pitch = angle

            point0 = camera2Fly(horizontal, pitch)
            point1 = camvec2Fly(x, y, horizontal, pitch, normalize=normalize_length)
            
            if yaw != 0:
                point0 = yaw_correct(
                        point0, yaw, horizontal, pitch)
                point1 = yaw_correct(
                        point1, yaw, horizontal, pitch)



            if correct_level:
                rotation = -self.antenna_level_correction
                point0 = rotate_about_x(point0, rotation)
                point1 = rotate_about_x(point1, rotation)
            
            x0,y0,z0 = point0
            x1,y1,z1 = point1

            #vectors.append( (tuple(angle), (x0,x1), (y0,y1), (z0, z1)) )
            points[i] = np.array(point0)            
            vectors[i] = np.array(point1) - points[i] 

        # Vertical/horizontal angle limiting
        booleans = where_vertical_between(points, lower=self.va_limits[0],
                upper=self.va_limits[1], reverse=self.alimits_reverse)
        points = points[booleans]
        vectors = vectors[booleans]

             
        if merge_distance:
            points, vectors = self._merge_by_distance(
                    points, vectors, merge_distance)

        if return_angles:
            return points, vectors, angles
        else:
            return points, vectors


    def get_recording_time(self, image_folder, i_rep=0):
        '''
        Returns the timestamp of a recording, if measure_movement() method has been
        run for the recording.

        If no time is found,
            returns None

        recording_Name      Recording name
        i_rep               Return the time for recording repeat i_rep
                                By default, i_rep=0
        '''

        angle = image_folder.lstrip('pos')

        for eye in ['left', 'right']:
            try:
                return self.movements[eye][angle][i_rep]['time']
            except (KeyError, IndexError):
                pass
            except AttributeError:
                print('No time for {} because movements not analysed'.format(image_folder))
                return None
        
        return None



    def stop(self):
        '''
        Stop long running activities (now measurement).
        '''
        self.stop_now = True


    # ------------
    # LINKED DATA
    # linking external data such as ERGs to the DPP data (MAnalyser)
    # ------------

    def link_data(self, key, data):
        '''
        Data linked to the MAnalyser
        '''
        self.linked_data[key] = data


    def save_linked_data(self):
        '''
        Attempt saving the linked data on disk in JSON format.
        '''
        os.makedirs(self._link_savedir, exist_ok=True)
        
        for key, data in self.linked_data.items():
            with open(os.path.join(self._link_savedir, "{}.json".format(key)), 'w') as fp:
                json.dump(data, fp)

    
    def load_linked_data(self):
        '''
        Load linked data from specimen datadir.
        '''
        # Initialize linked data to an empty dict
        self.linked_data = {}

        # Check if linked data directory exsists, if not, the no linked data for this specimen
        if os.path.exists(self._link_savedir):

            dfiles = [fn for fn in os.listdir(self._link_savedir) if fn.endswith('.json')]
            
            for dfile in dfiles:
                with open(os.path.join(self._link_savedir, dfile), 'r') as fp:
                    data = json.load(fp)
                    self.linked_data[dfile.replace('.json', '')] = data
        

class MAverager(AnalyserBase):
    '''
    Combining and averaging results from many MAnalyser objects.
    
    MAverager acts like MAnalyser object for getting data (like get_2d_vectors)
    but lacks the movement analysis (cross-correlation) related parts.
    '''
    def __init__(self, manalysers):
        super().__init__()

        self.manalysers = manalysers

        self.interpolation = {}

        self.intp_step = (5, 5)
        
        self.eyes = manalysers[0].eyes
        self.vector_rotation = manalysers[0].vector_rotation

        self.interpolated_raw= {}
        
        # Info data about available UI options
        self.ui_options = manalysers[0].ui_options.copy() 

    def get_N_specimens(self):
        return len(self.manalysers)


    def get_specimen_name(self):
        return 'averaged_'+'_'.join([manalyser.folder for manalyser in self.manalysers])
    
    @property
    def name(self):
        return self.get_specimen_name()
    
    @name.setter
    def name(self, name):
        return


    def setInterpolationSteps(self, horizontal_step, vertical_step):
        '''
        Set the resolution of the N-nearest neighbour interpolation in Maverager.get_2d_vectors.

        INPUT ARGUMENTS
        horizontal_step
        vertical_step

        Arguments horizontal_step and vertical_step refer to the rotation stages.

        '''

        self.intp_step = (horizontal_step, vertical_step)

    
    def get_2d_vectors(self, eye, **kwargs):
        '''
        Get's the 2D movement vectors (in the camera coordinate system)
        using N_nearest neighbour interpolation and averaging.
        '''
        #Modified from get_3d_vectors
        
        interpolated = [[],[],[]]
        
        points_2d = []
        vectors_2d = []

        for analyser in self.manalysers:
            angles, X, Y = analyser.get_2d_vectors(eye, mirror_horizontal=False, mirror_pitch=False)
            vecs = [[x,y] for x,y in zip(X, Y)]
            points_2d.append(np.array(angles))
            vectors_2d.append( np.array(vecs) )
        
        vectors_2d = np.array(vectors_2d)
        
        kdtrees = [KDTree(points) for points in points_2d]

        hmin, hmax = (-90, 90)
        vmax, hmax = (-180, 180)
       
        intp_points = []
        for h in np.arange(hmin, hmax+0.01, 10):
            for v in np.arange(hmin, hmax+0.1, 10):
                intp_points.append((h,v))
        
        for intp_point in intp_points:
            
            nearest_vectors = []

            for kdtree, vectors in zip(kdtrees, vectors_2d):
                distance, index = kdtree.query(intp_point)
                
                if distance < math.sqrt(self.intp_step[0]**2+self.intp_step[1]**2):
                    nearest_vectors.append(vectors[index])

            if len(nearest_vectors) > len(vectors_2d)/2:
                avec = np.mean(nearest_vectors, axis=0)
                avec /= np.linalg.norm(avec)
                interpolated[0].append(np.array(intp_point))
                interpolated[1].append(avec[0])
                interpolated[2].append(avec[1])

        angles, x, y = interpolated
        return angles, x, y


    def get_3d_vectors(self, eye, correct_level=True, normalize_length=0.1,
            recalculate=False, strict=False, vertical_hardborder=False,
            repeats_separately=False, **kwargs):
        '''
        Equivalent to MAnalysers get_3d_vectors but interpolates with N-nearest
        neughbours.

        repeats_separately : bool
            If True, return underlying MAnalyser vectors separetly
            (same points get repeated many times)
        '''

        cachename = ';'.join([str(item) for item in [self.vector_rotation, correct_level, normalize_length, strict, vertical_hardborder]])


        if self.interpolation.get(eye, {}).get(cachename) is None or recalculate:

            interpolated = [[],[]]
            self.interpolated_raw[eye] = [] # key points str, value list of vectors

            R = 1
            intp_dist = (2 * R * np.sin(math.radians(self.intp_step[0])))
            
            vectors_3d = []

            for analyser in self.manalysers:
                analyser.vector_rotation = self.vector_rotation
                vec = analyser.get_3d_vectors(eye, correct_level=True,
                        normalize_length=normalize_length, **kwargs)
                
                # Case no vectors (point is empty)
                if len(vec[0]) == 0:
                    continue

                vectors_3d.append(vec)
            
            if not strict:
                intp_points = optimal_sampling(np.arange(-90, 90.01, self.intp_step[0]), np.arange(0, 360.01, self.intp_step[1]))
            else:
                if eye == 'left':
                    intp_points = optimal_sampling(np.arange(-90, 0.01, self.intp_step[0]), np.arange(0, 360.01, self.intp_step[1]))
                else:
                    intp_points = optimal_sampling(np.arange(0, 90.01, self.intp_step[0]), np.arange(0, 360.01, self.intp_step[1]))

            for intp_point in intp_points:
                
                nearest_vectors = []
                for vectors in vectors_3d:
                    i_nearest = nearest_neighbour(intp_point, vectors[0], max_distance=intp_dist)
                    
                    if not i_nearest is False:
                        
                        if vertical_hardborder:
                            if np.sign(intp_point[2]) != np.sign(vectors[0][i_nearest][2]):
                                continue

                        nearest_vectors.append(vectors[1][i_nearest])

                if len(nearest_vectors) > len(vectors_3d)/2:
                    avec = mean_vector(intp_point, nearest_vectors)
                    interpolated[0].append(np.array(intp_point))
                    interpolated[1].append(avec)
                    
                    self.interpolated_raw[eye].append(nearest_vectors)

            self.interpolation[eye] = {}
            self.interpolation[eye][cachename] = np.array(interpolated[0]), np.array(interpolated[1])
            
        else:
            pass
        
        
        points, vectors = self.interpolation[eye][cachename]
        
        if repeats_separately:
            newpoints = []
            newvecs = []
            for i_point, point in enumerate(points):
                for vec in self.interpolated_raw[eye][i_point]:
                    newpoints.append(point)
                    newvecs.append(vec)
            points = np.array(newpoints)
            vectors = np.array(newvecs)

        
        # Vertical/horizontal angle limiting
        booleans = where_vertical_between(points, lower=self.va_limits[0],
                upper=self.va_limits[1], reverse=self.alimits_reverse)
        points = points[booleans]
        vectors = vectors[booleans]
        

        return points, vectors




'''
Code paths related to manually alinging the vertical angles for many
specimens (aka. antenna level or zero correction)
'''

import os
import ast

import numpy as np
import matplotlib.pyplot as plt

from gonioanalysis.directories import ANALYSES_SAVEDIR, CODE_ROOTDIR, PROCESSING_TEMPDIR
from gonioanalysis.droso import DrosoSelect
from gonioanalysis.drosom.loading import load_data
from gonioanalysis.image_tools import ImageShower
from gonioanalysis.binary_search import binary_search_middle
from gonioanalysis.rotary_encoders import to_degrees

ZERO_CORRECTIONS_SAVEDIR = os.path.join(PROCESSING_TEMPDIR, 'vectical_corrections')


def load_reference_fly(reference_name):
    '''
    Returns the reference fly data, dictionary with pitch angles as keys and
    image filenames as items.
    '''
    pitches = []
    with open(os.path.join(ZERO_CORRECTIONS_SAVEDIR, reference_name, 'pitch_angles.txt'), 'r') as fp:
        for line in fp:
            pitches.append(line)

    images = [os.path.join(ZERO_CORRECTIONS_SAVEDIR, reference_name, fn) for fn in os.listdir(
        os.path.join(ZERO_CORRECTIONS_SAVEDIR, reference_name)) if fn.endswith('.tif') or fn.endswith('.tiff')]

    images.sort() 
    return {pitch: fn for pitch,fn in zip(pitches, images)}



#OLD def _drosom_load(self, folder):
def load_drosom(folder):
    '''
    Loads frontal images of the specified drosom
    
    folder          Path to drosom folder
    '''

    data = load_data(folder)

    pitches =[]
    images = []

    for str_angle_pair in data.keys():
        #angle_pair = strToDegrees(str_angle_pair)
        i_start = str_angle_pair.index('(')
        i_end = str_angle_pair.index(')')+1

        #print(str_angle_pair[i_start:i_end])
        angle_pair = [list(ast.literal_eval(str_angle_pair[i_start:i_end]))]
        to_degrees(angle_pair)
        angle_pair = angle_pair[0]

        if -10 < angle_pair[0] < 10:
            pitches.append(angle_pair[1])
            images.append(data[str_angle_pair][0][0])
    
    pitches, images = zip(*sorted(zip(pitches, images)))

    return pitches, images


def save_antenna_level_correction(fly_name, result):
    '''
    Saves the antenna level correction that should be a float
    '''
    directory = ZERO_CORRECTIONS_SAVEDIR
    os.makedirs(directory, exist_ok=True)
    with open(os.path.join(directory, fly_name+'.txt'), 'w') as fp:
        fp.write(str(float(result)))
 

class AntennaLevelFinder:
    '''
    Ask user to find the antenna levels
    '''
    
    
    def find_level(self, folder):
        '''
        Call this to make user to select antenna levels.
        folder      Full path to the folder
        '''
        
        fly = os.path.split(folder)[1]
        
        #if os.path.exists(os.path.join(ANALYSES_SAVEDIR, 'antenna_levels', fly+'.txt')):
        #    print('Fly {} is already analysed. Redo (y/n)?'.format(fly))
        #    if not input('>> ').lower() in ['yes', 'y']:
        #        return False

        if 'DrosoX' in fly or 'DrosoALR' in fly:
            # If DrosoX, use drosox loader and find antenna levels by user
            # driven binary search. 
            
            fig, ax = plt.subplots()
            shower = ImageShower(fig, ax)
            
            if 'DrosoALR' in fly:
                arl = True
            else:
                arl = False

            pitches, images = self._drosox_load(folder, arl)
            
            shower.setImages(images)
            center = binary_search_middle(len(images), shower)
            
            shower.close()
            
            result = str(pitches[center])


        else:
        
            # DrosoM is harder, there's images every 10 degrees in pitch.
            # Solution: Find closest matches using analysed DrosoX data
            
            # Load DrosoM data
            pitches, images = self._drosom_load(folder)
            
            
            # Load reference fly data
            reference_pitches = {fn: pitch for pitch, fn in load_reference_fly('alr_data').items()}
            #print(reference_pitches)
            reference_images = list(reference_pitches.keys())
            reference_images.sort()

            fig1, ax1 = plt.subplots()
            fig1.canvas.set_window_title('Reference Drosophila')
            ref_shower = ImageShower(fig1, ax1)
            ref_shower.setImages(reference_images)
            
            fig2, ax2 = plt.subplots()
            fig2.canvas.set_window_title('{}'.format(fly))
            m_shower = ImageShower(fig2, ax2)
            
            offsets = []

            for pitch, image in zip(pitches, images):
                
                m_shower.setImages([image])
                m_shower.setImage(0)

                #best_drosox_image = matcher.findBest(image, drosox_images)
                best_drosoref_image = reference_images[ binary_search_middle(len(reference_images), ref_shower) ]
                 
                reference_pitch = float(reference_pitches[best_drosoref_image])
                
                print('Pitch {}, reference {}'.format(pitch, reference_pitch))

                offsets.append( pitch - reference_pitch)
            
            ref_shower.close()
            m_shower.close()

            result = np.mean(offsets)
            
            print('Reporting mean offset of {} (from {})'.format(result, offsets))


        with open(os.path.join(ANALYSES_SAVEDIR, 'antenna_levels', fly+'.txt'), 'w') as fp:
            fp.write(str(float(result)))
       
   
    
    def _load_drosox_reference(self, folder, fly):
        '''
        Load DrosoX reference for DrosoM antenna level finding.
        
        returns dictionary dict = {image_fn1: correted_pitch_1, ....}
        '''
        pitches, images = self._drosox_load(folder, True)
        
        with open(os.path.join(ZERO_CORRECTIONS_SAVEDIR, fly+'.txt'), 'r') as fp:
            offset = float(fp.read())

        return {image: pitch-offset for image, pitch in zip(images, pitches)}




    def _drosox_load(self, folder, arl):
        '''
        Private method, not intented to be called from outside.
        Loads pitches and images.
        '''

        xloader = XLoader()
        data = xloader.getData(folder, arl_fly=arl)

        pitches = []
        images = []


 
        # Try to open if any previously analysed data
        analysed_data = []
        try: 
            with open(os.path.join(ANALYSES_SAVEDIR, 'binary_search', 'results_{}.json'.format(fly)), 'r') as fp:
                analysed_data = json.load(fp)
        except:
            pass
        analysed_pitches = [item['pitch'] for item in analysed_data]
        

        for i, (pitch, hor_im) in enumerate(data):
            
            if pitch in analysed_pitches:
                j = [k for k in range(len(analysed_pitches)) if analysed_pitches[k]['pitch'] == pitch]
                center_index = analysed_data[j]['index_middle']
            else:
                center_index = int(len(hor_im)/2)
            
            pitches.append(pitch)
            images.append(hor_im[center_index][1])
            
        return pitches, images




def main():
    finder = AntennaLevelFinder()
    
    selector = DrosoSelect()
    folders = selector.askUser()
       
    for folder in folders:
        finder.find_level(folder)

if __name__ == "__main__":
    # FIXME TODO
    main()



'''
Add grid lines to the selected GHS-DPP images.
Usefull for example still-image series illustrations.
'''
import os

import numpy as np
from tkinter import filedialog, simpledialog
import cv2

def main():

    fns = filedialog.askopenfilenames()
   
    if fns:

        print(fns)

        pixel_size = 0.817
        every = 10

        directory = os.path.dirname(fns[0])
        
        newdir = os.path.join(directory, 'gridded')
        os.makedirs(newdir, exist_ok=True)
        
        for fn in fns:
            image = cv2.imread(fn)
            

            for i_line, j in enumerate(np.arange(0, image.shape[0]-1, every/pixel_size)):
                j = int(j)
                
                if True:    
                    image[j:j+1, :, 2] = 255
                    image[j:j+1, :, 0] = 0
                    image[j:j+1, :, 1] = 0
                else:
                    image[j:j+1, :, 0] = 255
                    image[j:j+1, :, 1] = 0
                    image[j:j+1, :, 2] = 0
                    
            
            for i_line, i in enumerate(np.arange(0, image.shape[1]-1, every/pixel_size)):
                i = int(i)

                if True: 
                    image[:, i:i+1, 2] = 255
                    image[:, i:i+1, 0] = 0
                    image[:, i:i+1, 1] = 0
                else:
                    image[:, i:i+1, 0] = 255
                    image[:, i:i+1, 1] = 0
                    image[:, i:i+1, 2] = 0
                    
            cv2.imwrite(os.path.join(newdir, os.path.basename(fn)), image)


if __name__ == "__main__":
    main()


